import logging
import os
from traceback import format_exception

from django.conf import settings

from celery import Celery
from celery.signals import setup_logging, task_failure

from .util.core_helpers import get_site_preferences
from .util.email import send_email

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "aleksis.core.settings")

app = Celery("aleksis")  # noqa
app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks()


@task_failure.connect
def task_failure_notifier(
    sender=None, task_id=None, exception=None, args=None, kwargs=None, traceback=None, **__
):
    recipient_list = [e[1] for e in settings.ADMINS]
    send_email(
        template_name="celery_failure",
        from_email=get_site_preferences()["mail__address"],
        recipient_list=recipient_list,
        context={
            "task_name": sender.name,
            "task": str(sender),
            "task_id": str(task_id),
            "exception": str(exception),
            "args": args,
            "kwargs": kwargs,
            "traceback": "".join(format_exception(type(exception), exception, traceback)),
        },
    )


@setup_logging.connect
def on_setup_logging(*args, **kwargs):
    """Load Django's logging configuration when running inside Celery."""
    logging.config.dictConfig(settings.LOGGING)



from functools import wraps


def pwa_cache(view_func):
    """Add headers to a response so that the PWA will recognize it as cacheable."""

    @wraps(view_func)
    def _wrapped_view_func(request, *args, **kwargs):
        # Ensure argument looks like a request.
        if not hasattr(request, "META"):
            raise TypeError(
                "pwa_cache didn't receive an HttpRequest. If you are "
                "decorating a classmethod, be sure to use @method_decorator."
            )
        response = view_func(request, *args, **kwargs)
        response.headers["PWA-Is-Cacheable"] = "true"
        response.headers["Access-Control-Expose-Headers"] = "PWA-Is-Cacheable"
        return response

    return _wrapped_view_func



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0056_rename_customevent_personalevent"),
    ]

    operations = [
        migrations.RunSQL(
            "; ".join(
                f"drop table if exists {x} cascade;" for x in [
                    "otp_yubikey_remoteyubikeydevice",
                    "otp_yubikey_validationservice",
                    "otp_yubikey_remoteyubikeydevice"
                ]
            )
        )
    ]



# Generated by Django 3.2.10 on 2021-12-25 10:59

import aleksis.core.mixins
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import migrations, models
import django.db.models.deletion


def assume_ldap_authenticated_true(apps, schema_editor):
    """Set ldap_authenticated user attribute to True to protect existing sites."""
    if not hasattr(settings, "AUTH_LDAP_SERVER_URI"):
        # Skip if LDAP is not used on site
        return

    User = get_user_model()
    UserAdditionalAttributes = apps.get_model("core", "UserAdditionalAttributes")

    db_alias = schema_editor.connection.alias

    attributes = [
        UserAdditionalAttributes(user_id=user.pk, attributes={"ldap_authenticated": True})
        for user in User.objects.using(db_alias).all()
    ]
    UserAdditionalAttributes.objects.using(db_alias).bulk_create(attributes)


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0029_invitations'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserAdditionalAttributes',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attributes', models.JSONField(default=dict, verbose_name='Additional attributes')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='additional_attributes', to=settings.AUTH_USER_MODEL, verbose_name='Linked user')),
            ],
            bases=(models.Model, aleksis.core.mixins.PureDjangoModel),
        ),
        migrations.RunPython(assume_ldap_authenticated_true, lambda a, s: None),
    ]



# Generated by Django 3.2.11 on 2022-01-08 10:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0031_oauthapplication_icon'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='person',
            name='is_active',
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-22 18:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0058_migrate_preferences_to_global"),
    ]

    operations = ([
        migrations.AlterModelManagers(
            name="activity",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="additionalfield",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="announcementrecipient",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="custommenu",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="custommenuitem",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="datacheckresult",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="grouptype",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="notification",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="organisation",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="pdffile",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="person",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="persongroupthrough",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="room",
            managers=[],
        ),
        migrations.AlterModelManagers(
            name="taskuserassignment",
            managers=[],
        ),
        migrations.RemoveConstraint(
            model_name="additionalfield",
            name="unique_title_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="custommenu",
            name="unique_menu_name_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="group",
            name="unique_school_term_name",
        ),
        migrations.RemoveConstraint(
            model_name="group",
            name="unique_school_term_short_name",
        ),
        migrations.RemoveConstraint(
            model_name="grouptype",
            name="unique_group_type_name_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="person",
            name="unique_short_name_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="room",
            name="unique_room_short_name_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="schoolterm",
            name="unique_school_term_name_per_site",
        ),
        migrations.RemoveConstraint(
            model_name="schoolterm",
            name="unique_school_term_dates_per_site",
        ),
    ] + [
        migrations.RunSQL(
            f"ALTER TABLE core_{model_name} drop column if exists site_id;"
        ) for model_name  in
        [
            "activity",
            "additionalfield",
            "announcement",
            "announcementrecipient",
            "calendarevent",
            "custommenu",
            "custommenuitem",
            "dashboardwidgetorder",
            "datacheckresult",
            "group",
            "grouptype",
            "notification",
            "organisation",
            "pdffile",
            "person",
            "persongroupthrough",
            "room",
            "schoolterm",
            "taskuserassignment"
        ]
    ] +  [
        migrations.DeleteModel(
            name="SitePreferenceModel",
        ),
        migrations.AddConstraint(
            model_name="additionalfield",
            constraint=models.UniqueConstraint(fields=("title",), name="unique_title"),
        ),
        migrations.AddConstraint(
            model_name="custommenu",
            constraint=models.UniqueConstraint(fields=("name",), name="unique_menu_name"),
        ),
        migrations.AddConstraint(
            model_name="group",
            constraint=models.UniqueConstraint(
                fields=("school_term", "name"), name="unique_school_term_name_group"
            ),
        ),
        migrations.AddConstraint(
            model_name="group",
            constraint=models.UniqueConstraint(
                fields=("school_term", "short_name"), name="unique_school_term_short_name_group"
            ),
        ),
        migrations.AddConstraint(
            model_name="grouptype",
            constraint=models.UniqueConstraint(fields=("name",), name="unique_group_type_name"),
        ),
        migrations.AddConstraint(
            model_name="person",
            constraint=models.UniqueConstraint(fields=("short_name",), name="unique_short_name"),
        ),
        migrations.AddConstraint(
            model_name="room",
            constraint=models.UniqueConstraint(
                fields=("short_name",), name="unique_room_short_name"
            ),
        ),
        migrations.AddConstraint(
            model_name="schoolterm",
            constraint=models.UniqueConstraint(fields=("name",), name="unique_school_term_name"),
        ),
        migrations.AddConstraint(
            model_name="schoolterm",
            constraint=models.UniqueConstraint(
                fields=("date_start", "date_end"), name="unique_school_term_dates"
            ),
        ),
    ])



# Generated by Django 4.1.7 on 2023-03-23 18:30

from django.db import migrations, models
import django.db.models.deletion
import aleksis.core.managers


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0053_freebusy"),
    ]

    operations = [
        migrations.CreateModel(
            name="Organisation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                ("email", models.CharField(max_length=255, verbose_name="Email")),
                (
                    "related_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.group",
                        verbose_name="Related group",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            managers=[
            ],
        ),
        migrations.AlterModelOptions(
            name="organisation",
            options={"verbose_name": "Organisation", "verbose_name_plural": "Organisations"},
        ),
        migrations.AlterModelManagers(
            name="organisation",
            managers=[
                ("objects", aleksis.core.managers.AlekSISBaseManager()),
            ],
        ),
        migrations.AddField(
            model_name="organisation",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
    ]



import rules
from rules import is_superuser

from .models import Announcement, Group, GroupType, Holiday, Person
from .util.predicates import (
    has_any_object,
    has_global_perm,
    has_object_perm,
    has_person,
    is_anonymous,
    is_current_person,
    is_group_owner,
    is_notification_recipient,
    is_own_celery_task,
    is_personal_event_owner,
    is_site_preference_set,
)

rules.add_perm("core", rules.always_allow)

# Login
login_predicate = is_anonymous
rules.add_perm("core.login_rule", login_predicate)

# Logout
logout_predicate = ~is_anonymous
rules.add_perm("core.logout_rule", logout_predicate)

# Account
view_account_predicate = has_person
rules.add_perm("core.view_account_rule", view_account_predicate)

# 2FA
manage_2fa_predicate = has_person
rules.add_perm("core.manage_2fa_rule", manage_2fa_predicate)

# Social Connections
manage_social_connections_predicate = has_person
rules.add_perm("core.manage_social_connections_rule", manage_social_connections_predicate)

# Authorized tokens
manage_authorized_tokens_predicate = has_person
rules.add_perm("core.manage_authorized_tokens_rule", manage_authorized_tokens_predicate)

# View dashboard
view_dashboard_predicate = is_site_preference_set("general", "anonymous_dashboard") | has_person
rules.add_perm("core.view_dashboard_rule", view_dashboard_predicate)

# View notifications
rules.add_perm("core.view_notifications_rule", has_person)

# Use search
search_predicate = has_person & has_global_perm("core.search")
rules.add_perm("core.search_rule", search_predicate)

# View persons
view_persons_predicate = has_person & (
    has_global_perm("core.view_person") | has_any_object("core.view_person", Person)
)
rules.add_perm("core.view_persons_rule", view_persons_predicate)

# View person
view_person_predicate = has_person & (
    is_current_person | has_global_perm("core.view_person") | has_object_perm("core.view_person")
)
rules.add_perm("core.view_person_rule", view_person_predicate)

# View person address
view_address_predicate = has_person & (
    is_current_person | has_global_perm("core.view_address") | has_object_perm("core.view_address")
)
rules.add_perm("core.view_address_rule", view_address_predicate)

# View person contact details
view_contact_details_predicate = has_person & (
    is_current_person
    | has_global_perm("core.view_contact_details")
    | has_object_perm("core.view_contact_details")
)
rules.add_perm("core.view_contact_details_rule", view_contact_details_predicate)

# View person photo
view_photo_predicate = has_person & (
    is_current_person | has_global_perm("core.view_photo") | has_object_perm("core.view_photo")
)
rules.add_perm("core.view_photo_rule", view_photo_predicate)

# View person avatar image
view_avatar_predicate = has_person & (
    is_current_person | has_global_perm("core.view_avatar") | has_object_perm("core.view_avatar")
)
rules.add_perm("core.view_avatar_rule", view_avatar_predicate)

# View persons groups
view_groups_predicate = has_person & (
    is_current_person
    | has_global_perm("core.view_person_groups")
    | has_object_perm("core.view_person_groups")
)
rules.add_perm("core.view_person_groups_rule", view_groups_predicate)

# Edit person
edit_person_predicate = has_person & (
    is_current_person & is_site_preference_set("account", "editable_fields_person")
    | has_global_perm("core.change_person")
    | has_object_perm("core.change_person")
)
rules.add_perm("core.edit_person_rule", edit_person_predicate)

# Delete person
delete_person_predicate = has_person & (
    has_global_perm("core.delete_person") | has_object_perm("core.delete_person")
)
rules.add_perm("core.delete_person_rule", delete_person_predicate)

# View groups
view_groups_predicate = has_person & (
    has_global_perm("core.view_group") | has_any_object("core.view_group", Group)
)
rules.add_perm("core.view_groups_rule", view_groups_predicate)

# View group
view_group_predicate = has_person & (
    has_global_perm("core.view_group") | has_object_perm("core.view_group")
)
rules.add_perm("core.view_group_rule", view_group_predicate)

# Edit group
edit_group_predicate = has_person & (
    has_global_perm("core.change_group") | has_object_perm("core.change_group")
)
rules.add_perm("core.edit_group_rule", edit_group_predicate)

# Delete group
delete_group_predicate = has_person & (
    has_global_perm("core.delete_group") | has_object_perm("core.delete_group")
)
rules.add_perm("core.delete_group_rule", delete_group_predicate)

# Assign child groups to groups
assign_child_groups_to_groups_predicate = has_person & has_global_perm(
    "core.assign_child_groups_to_groups"
)
rules.add_perm("core.assign_child_groups_to_groups_rule", assign_child_groups_to_groups_predicate)

# Edit school information
edit_school_information_predicate = has_person & has_global_perm("core.change_school")
rules.add_perm("core.edit_school_information_rule", edit_school_information_predicate)

# Manage data
manage_data_predicate = has_person & has_global_perm("core.manage_data")
rules.add_perm("core.manage_data_rule", manage_data_predicate)

# Mark notification as read
mark_notification_as_read_predicate = has_person & is_notification_recipient
rules.add_perm("core.mark_notification_as_read_rule", mark_notification_as_read_predicate)

# View announcements
view_announcements_predicate = has_person & (
    has_global_perm("core.view_announcement")
    | has_any_object("core.view_announcement", Announcement)
)
rules.add_perm("core.view_announcements_rule", view_announcements_predicate)

# Create or edit announcement
create_or_edit_announcement_predicate = has_person & (
    has_global_perm("core.add_announcement")
    & (has_global_perm("core.change_announcement") | has_object_perm("core.change_announcement"))
)
rules.add_perm("core.create_or_edit_announcement_rule", create_or_edit_announcement_predicate)

# Delete announcement
delete_announcement_predicate = has_person & (
    has_global_perm("core.delete_announcement") | has_object_perm("core.delete_announcement")
)
rules.add_perm("core.delete_announcement_rule", delete_announcement_predicate)

# Use impersonate
impersonate_predicate = has_person & has_global_perm("core.impersonate")
rules.add_perm("core.impersonate_rule", impersonate_predicate)

# View system status
view_system_status_predicate = has_person & has_global_perm("core.view_system_status")
rules.add_perm("core.view_system_status_rule", view_system_status_predicate)

# View people menu (persons + objects)
rules.add_perm(
    "core.view_people_menu_rule",
    has_person
    & (view_persons_predicate | view_groups_predicate | assign_child_groups_to_groups_predicate),
)

# View person personal details
view_personal_details_predicate = has_person & (
    is_current_person
    | has_global_perm("core.view_personal_details")
    | has_object_perm("core.view_personal_details")
)
rules.add_perm("core.view_personal_details_rule", view_personal_details_predicate)

# Change site preferences
change_site_preferences = has_person & (
    has_global_perm("core.change_site_preferences")
    | has_object_perm("core.change_site_preferences")
)
rules.add_perm("core.change_site_preferences_rule", change_site_preferences)

# Change person preferences
change_person_preferences = has_person & (
    is_current_person
    | has_global_perm("core.change_person_preferences")
    | has_object_perm("core.change_person_preferences")
)
rules.add_perm("core.change_person_preferences_rule", change_person_preferences)

# Change account preferences
change_account_preferences = has_person
rules.add_perm("core.change_account_preferences_rule", change_account_preferences)

# Change group preferences
change_group_preferences = has_person & (
    has_global_perm("core.change_group_preferences")
    | has_object_perm("core.change_group_preferences")
    | is_group_owner
)
rules.add_perm("core.change_group_preferences_rule", change_group_preferences)


# View group type
view_group_type_predicate = has_person & (
    has_global_perm("core.view_grouptype") | has_object_perm("core.view_grouptype")
)
rules.add_perm("core.view_grouptype_rule", view_group_type_predicate)

# Edit group type
change_group_type_predicate = has_person & (
    has_global_perm("core.change_grouptype") | has_object_perm("core.change_grouptype")
)
rules.add_perm("core.edit_grouptype_rule", change_group_type_predicate)

# Create group type
create_group_type_predicate = has_person & (
    has_global_perm("core.add_grouptype") | has_object_perm("core.add_grouptype")
)
rules.add_perm("core.create_grouptype_rule", create_group_type_predicate)


# Delete group type
delete_group_type_predicate = has_person & (
    has_global_perm("core.delete_grouptype") | has_object_perm("core.delete_grouptype")
)
rules.add_perm("core.delete_grouptype_rule", delete_group_type_predicate)

# View group types
view_group_types_predicate = has_person & (
    has_global_perm("core.view_grouptype") | has_any_object("core.view_grouptype", GroupType)
)
rules.add_perm("core.view_grouptypes_rule", view_group_types_predicate)

# Create person
create_person_predicate = has_person & (
    has_global_perm("core.add_person") | has_object_perm("core.add_person")
)
rules.add_perm("core.create_person_rule", create_person_predicate)

# Create group
create_group_predicate = has_person & (
    has_global_perm("core.add_group") | has_object_perm("core.add_group")
)
rules.add_perm("core.create_group_rule", create_group_predicate)

# School years
view_school_term_predicate = has_person & has_global_perm("core.view_schoolterm")
rules.add_perm("core.view_schoolterm_rule", view_school_term_predicate)

create_school_term_predicate = has_person & has_global_perm("core.add_schoolterm")
rules.add_perm("core.create_schoolterm_rule", create_school_term_predicate)

edit_school_term_predicate = has_person & has_global_perm("core.change_schoolterm")
rules.add_perm("core.edit_schoolterm_rule", edit_school_term_predicate)

# View group stats
view_group_stats_predicate = has_person & (
    has_global_perm("core.view_group_stats") | has_object_perm("core.view_group_stats")
)
rules.add_perm("core.view_group_stats_rule", view_group_stats_predicate)

# View data check results
view_data_check_results_predicate = has_person & has_global_perm("core.view_datacheckresult")
rules.add_perm("core.view_datacheckresults_rule", view_data_check_results_predicate)

# Run data checks
run_data_checks_predicate = (
    has_person & view_data_check_results_predicate & has_global_perm("core.run_data_checks")
)
rules.add_perm("core.run_data_checks_rule", run_data_checks_predicate)

# Solve data problems
solve_data_problem_predicate = (
    has_person & view_data_check_results_predicate & has_global_perm("core.solve_data_problem")
)
rules.add_perm("core.solve_data_problem_rule", solve_data_problem_predicate)

view_dashboard_widget_predicate = has_person & has_global_perm("core.view_dashboardwidget")
rules.add_perm("core.view_dashboardwidget_rule", view_dashboard_widget_predicate)

create_dashboard_widget_predicate = has_person & has_global_perm("core.add_dashboardwidget")
rules.add_perm("core.create_dashboardwidget_rule", create_dashboard_widget_predicate)

edit_dashboard_widget_predicate = has_person & has_global_perm("core.change_dashboardwidget")
rules.add_perm("core.edit_dashboardwidget_rule", edit_dashboard_widget_predicate)

delete_dashboard_widget_predicate = has_person & has_global_perm("core.delete_dashboardwidget")
rules.add_perm("core.delete_dashboardwidget_rule", delete_dashboard_widget_predicate)

edit_dashboard_predicate = is_site_preference_set("general", "dashboard_editing") & has_person
rules.add_perm("core.edit_dashboard_rule", edit_dashboard_predicate)

edit_default_dashboard_predicate = has_person & has_global_perm("core.edit_default_dashboard")
rules.add_perm("core.edit_default_dashboard_rule", edit_default_dashboard_predicate)

# django-allauth
signup_predicate = is_site_preference_set(section="auth", pref="signup_enabled")
rules.add_perm("core.signup_rule", signup_predicate)

change_password_predicate = has_person & is_site_preference_set(
    section="auth", pref="allow_password_change"
)
rules.add_perm("core.change_password_rule", change_password_predicate)

reset_password_predicate = is_site_preference_set(section="auth", pref="allow_password_reset")
rules.add_perm("core.reset_password_rule", reset_password_predicate)

# django-invitations
invite_enabled_predicate = is_site_preference_set(section="auth", pref="invite_enabled")
rules.add_perm("core.invite_enabled", invite_enabled_predicate)

accept_invite_predicate = has_person & invite_enabled_predicate
rules.add_perm("core.accept_invite_rule", accept_invite_predicate)

invite_predicate = has_person & invite_enabled_predicate & has_global_perm("core.invite")
rules.add_perm("core.invite_rule", invite_predicate)

# OAuth2 permissions
create_oauthapplication_predicate = has_person & has_global_perm("core.add_oauthapplication")
rules.add_perm("core.create_oauthapplication_rule", create_oauthapplication_predicate)

view_oauth_applications_predicate = has_person & has_global_perm("core.view_oauthapplication")
rules.add_perm("core.view_oauthapplications_rule", view_oauth_applications_predicate)

view_oauth_application_predicate = has_person & has_global_perm("core.view_oauthapplication")
rules.add_perm("core.view_oauthapplication_rule", view_oauth_application_predicate)

edit_oauth_application_predicate = has_person & has_global_perm("core.change_oauthapplication")
rules.add_perm("core.edit_oauthapplication_rule", edit_oauth_application_predicate)

delete_oauth_applications_predicate = has_person & has_global_perm("core.delete_oauth_applications")
rules.add_perm("core.delete_oauth_applications_rule", delete_oauth_applications_predicate)

view_django_admin_predicate = has_person & is_superuser
rules.add_perm("core.view_django_admin_rule", view_django_admin_predicate)

# View admin menu
view_admin_menu_predicate = has_person & (
    manage_data_predicate
    | view_school_term_predicate
    | impersonate_predicate
    | view_system_status_predicate
    | view_announcements_predicate
    | view_data_check_results_predicate
    | view_oauth_applications_predicate
    | view_dashboard_widget_predicate
    | view_django_admin_predicate
)
rules.add_perm("core.view_admin_menu_rule", view_admin_menu_predicate)

# Upload and browse files via CKEditor
upload_files_ckeditor_predicate = has_person & has_global_perm("core.upload_files_ckeditor")
rules.add_perm("core.upload_files_ckeditor_rule", upload_files_ckeditor_predicate)

manage_person_permissions_predicate = has_person & is_superuser
rules.add_perm("core.manage_permissions_rule", manage_person_permissions_predicate)

test_pdf_generation_predicate = has_person & has_global_perm("core.test_pdf")
rules.add_perm("core.test_pdf_rule", test_pdf_generation_predicate)

view_progress_predicate = has_person & is_own_celery_task
rules.add_perm("core.view_progress_rule", view_progress_predicate)

view_calendar_feed_predicate = has_person
rules.add_perm("core.view_calendar_feed_rule", view_calendar_feed_predicate)

# Holidays

view_holiday_predicate = has_person & (
    has_global_perm("core.view_holiday") | has_object_perm("core.view_holiday")
)
rules.add_perm("core.view_holiday_rule", view_holiday_predicate)

view_holidays_predicate = has_person & (
    has_global_perm("core.view_holiday") | has_any_object("core.view_holiday", Holiday)
)
rules.add_perm("core.view_holidays_rule", view_holidays_predicate)

edit_holiday_predicate = has_person & (
    has_global_perm("core.change_holiday") | has_object_perm("core.change_holiday")
)
rules.add_perm("core.edit_holiday_rule", edit_holiday_predicate)

create_holiday_predicate = has_person & (has_global_perm("core.add_holiday"))
rules.add_perm("core.create_holiday_rule", create_holiday_predicate)

delete_holiday_predicate = has_person & (
    has_global_perm("core.delete_holiday") | has_object_perm("core.delete_holiday")
)
rules.add_perm("core.delete_holiday_rule", delete_holiday_predicate)

# Custom events

create_personal_event_predicate = has_person
rules.add_perm("core.create_personal_event_rule", create_personal_event_predicate)

create_personal_event_with_invitations_predicate = has_person & has_global_perm(
    "core.create_personalevent"
)
rules.add_perm(
    "core.create_personal_event_with_invitations_rule",
    create_personal_event_with_invitations_predicate,
)

edit_personal_event_predicate = has_person & (
    has_global_perm("core.change_personalevent") | is_personal_event_owner
)
rules.add_perm("core.edit_personal_event_rule", edit_personal_event_predicate)

delete_personal_event_predicate = has_person & (
    has_global_perm("core.delete_personalevent") | is_personal_event_owner
)
rules.add_perm("core.delete_personal_event_rule", delete_personal_event_predicate)



from graphene_django import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
)
from guardian.shortcuts import get_objects_for_user

from ..models import Holiday
from .base import (
    DjangoFilterMixin,
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
    PermissionsTypeMixin,
)


class HolidayType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Holiday
        fields = ("id", "holiday_name", "date_start", "date_end")
        filter_fields = {
            "id": ["exact", "lte", "gte"],
            "holiday_name": ["icontains"],
            "date_start": ["exact", "lte", "gte"],
            "date_end": ["exact", "lte", "gte"],
        }

    @classmethod
    def get_queryset(cls, queryset, info):
        return get_objects_for_user(info.context.user, "core.view_holiday", queryset)


class HolidayBatchCreateMutation(DjangoBatchCreateMutation):
    class Meta:
        model = Holiday
        permissions = ("core.create_holiday_rule",)
        only_fields = ("holiday_name", "date_start", "date_end")


class HolidayBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = Holiday
        permissions = ("core.delete_holiday_rule",)


class HolidayBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = Holiday
        permissions = ("core.edit_holiday_rule",)
        only_fields = ("id", "holiday_name", "date_start", "date_end")



from typing import TYPE_CHECKING, Any, Optional

import django.apps
from django.conf import settings
from django.contrib import messages
from django.http import HttpRequest
from django.utils.module_loading import autodiscover_modules
from django.utils.translation import gettext as _

from dynamic_preferences.registries import preference_models
from health_check.plugins import plugin_dir
from oauthlib.common import Request as OauthlibRequest

from .registries import group_preferences_registry, person_preferences_registry
from .util.apps import AppConfig
from .util.core_helpers import (
    create_default_celery_schedule,
    get_or_create_favicon,
    get_site_preferences,
    has_person,
)
from .util.sass_helpers import clean_scss

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class CoreConfig(AppConfig):
    name = "aleksis.core"
    verbose_name = "AlekSIS — The Free School Information System"
    dist_name = "AlekSIS-Core"

    urls = {
        "Repository": "https://edugit.org/AlekSIS/official/AlekSIS/",
    }
    licence = "EUPL-1.2+"
    copyright_info = (
        ([2017, 2018, 2019, 2020, 2021, 2022, 2023], "Jonathan Weth", "wethjo@katharineum.de"),
        ([2017, 2018, 2019, 2020], "Frank Poetzsch-Heffter", "p-h@katharineum.de"),
        ([2018, 2019, 2020, 2021, 2022, 2023], "Hangzhi Yu", "yuha@katharineum.de"),
        ([2018, 2019, 2020, 2021, 2022, 2023], "Julian Leucker", "leuckeju@katharineum.de"),
        ([2019, 2020, 2021, 2022, 2023], "Dominik George", "dominik.george@teckids.org"),
        ([2019, 2020, 2021, 2022], "Tom Teichler", "tom.teichler@teckids.org"),
        ([2019], "mirabilos", "thorsten.glaser@teckids.org"),
        ([2021, 2022, 2023], "magicfelix", "felix@felix-zauberer.de"),
        ([2021], "Lloyd Meins", "meinsll@katharineum.de"),
        ([2022], "Benedict Suska", "benedict.suska@teckids.org"),
        ([2022], "Lukas Weichelt", "lukas.weichelt@teckids.org"),
    )

    def ready(self):
        super().ready()

        from django.conf import settings  # noqa

        # Autodiscover various modules defined by AlekSIS
        autodiscover_modules("model_extensions", "form_extensions", "checks")

        personpreferencemodel = self.get_model("PersonPreferenceModel")
        grouppreferencemodel = self.get_model("GroupPreferenceModel")

        preference_models.register(personpreferencemodel, person_preferences_registry)
        preference_models.register(grouppreferencemodel, group_preferences_registry)

        from .health_checks import (
            BackupJobHealthCheck,
            DataChecksHealthCheckBackend,
            DbBackupAgeHealthCheck,
            MediaBackupAgeHealthCheck,
        )

        plugin_dir.register(DataChecksHealthCheckBackend)
        plugin_dir.register(DbBackupAgeHealthCheck)
        plugin_dir.register(MediaBackupAgeHealthCheck)
        plugin_dir.register(BackupJobHealthCheck)

    def preference_updated(
        self,
        sender: Any,
        section: Optional[str] = None,
        name: Optional[str] = None,
        old_value: Optional[Any] = None,
        new_value: Optional[Any] = None,
        **kwargs,
    ) -> None:
        from django.conf import settings  # noqa

        if section == "theme":
            if name in ("primary", "secondary"):
                clean_scss()
            elif name in ("favicon", "pwa_icon"):
                from favicon.models import Favicon, FaviconImg  # noqa

                is_favicon = name == "favicon"

                if new_value:
                    # Get file object from preferences instead of using new_value
                    # to prevent problems with special file storages
                    file_obj = get_site_preferences()[f"{section}__{name}"]

                    favicon = Favicon.objects.update_or_create(
                        title=name,
                        defaults={"isFavicon": is_favicon, "faviconImage": file_obj},
                    )[0]
                    FaviconImg.objects.filter(faviconFK=favicon).delete()
                else:
                    Favicon.objects.filter(title=name, isFavicon=is_favicon).delete()
                    if name in settings.DEFAULT_FAVICON_PATHS:
                        get_or_create_favicon(
                            name, settings.DEFAULT_FAVICON_PATHS[name], is_favicon=is_favicon
                        )

    def pre_migrate(
        self,
        app_config: django.apps.AppConfig,
        verbosity: int,
        interactive: bool,
        using: str,
        plan: list[tuple],
        apps: django.apps.registry.Apps,
        **kwargs,
    ) -> None:
        super().pre_migrate(app_config, verbosity, interactive, using, plan, apps)
        from .data_checks import check_data_for_migrations

        # Run data checks to validate data
        check_data_for_migrations(with_dependencies=True)

    def post_migrate(
        self,
        app_config: django.apps.AppConfig,
        verbosity: int,
        interactive: bool,
        using: str,
        **kwargs,
    ) -> None:
        from django.conf import settings  # noqa
        from .data_checks import check_data_for_migrations

        super().post_migrate(app_config, verbosity, interactive, using, **kwargs)

        # Ensure that default Favicon object exists
        for name, default in settings.DEFAULT_FAVICON_PATHS.items():
            get_or_create_favicon(name, default, is_favicon=name == "favicon")

        # Create default periodic tasks
        create_default_celery_schedule()

        # Run data checks to validate data
        check_data_for_migrations()

    def user_logged_in(
        self, sender: type, request: Optional[HttpRequest], user: "User", **kwargs
    ) -> None:
        if has_person(user):
            # Save the associated person to pick up defaults
            user.person.save()

    def user_logged_out(
        self, sender: type, request: Optional[HttpRequest], user: "User", **kwargs
    ) -> None:
        messages.success(request, _("You have been logged out successfully."))

    @classmethod
    def get_all_scopes(cls) -> dict[str, str]:
        scopes = {
            "read": "Read anything the resource owner can read",
            "write": "Write anything the resource owner can write",
        }
        if settings.OAUTH2_PROVIDER.get("OIDC_ENABLED", False):
            scopes |= {
                "openid": _("OpenID Connect scope"),
                "profile": _("Given name, family name, link to profile and picture if existing."),
                "address": _("Full home postal address"),
                "email": _("Email address"),
                "phone": _("Home and mobile phone"),
                "groups": _("Groups"),
            }
        return scopes

    @classmethod
    def get_additional_claims(cls, scopes: list[str], request: OauthlibRequest) -> dict[str, Any]:
        django_request = HttpRequest()
        django_request.META = request.headers

        claims = {
            "preferred_username": request.user.username,
        }

        if "profile" in scopes:
            if has_person(request.user):
                claims["given_name"] = request.user.person.first_name
                claims["family_name"] = request.user.person.last_name
                claims["profile"] = django_request.build_absolute_uri(
                    request.user.person.get_absolute_url()
                )
                if request.user.person.avatar:
                    claims["picture"] = django_request.build_absolute_uri(
                        request.user.person.avatar.url
                    )
            else:
                claims["given_name"] = request.user.first_name
                claims["family_name"] = request.user.last_name

        if "email" in scopes:
            if has_person(request.user):
                claims["email"] = request.user.person.email
            else:
                claims["email"] = request.user.email

        if "address" in scopes and has_person(request.user):
            claims["address"] = {
                "street_address": request.user.person.street
                + " "
                + request.user.person.housenumber,
                "locality": request.user.person.place,
                "postal_code": request.user.person.postal_code,
            }

        if "phone" in scopes and has_person(request.user):
            claims["mobile_number"] = request.user.person.mobile_number
            claims["phone_number"] = request.user.person.phone_number

        if "groups" in scopes and has_person(request.user):
            claims["groups"] = list(
                request.user.person.member_of.values_list("name", flat=True).all()
            )

        return claims



# Generated by Django 4.1.10 on 2023-07-11 19:01

import aleksis.core.managers
import aleksis.core.mixins
from django.db import migrations, models
import django.db.models.deletion
import recurrence.fields
import timezone_field.fields


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("core", "0050_managed_by_app_label"),
    ]

    operations = [
        migrations.CreateModel(
            name="CalendarEvent",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "managed_by_app_label",
                    models.CharField(
                        blank=True,
                        editable=False,
                        max_length=255,
                        verbose_name="App label of app responsible for managing this instance",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                (
                    "datetime_start",
                    models.DateTimeField(blank=True, null=True, verbose_name="Start date and time"),
                ),
                (
                    "datetime_end",
                    models.DateTimeField(blank=True, null=True, verbose_name="End date and time"),
                ),
                (
                    "timezone",
                    timezone_field.fields.TimeZoneField(
                        blank=True, null=True, verbose_name="Timezone"
                    ),
                ),
                ("date_start", models.DateField(blank=True, null=True, verbose_name="Start date")),
                ("date_end", models.DateField(blank=True, null=True, verbose_name="End date")),
                (
                    "recurrences",
                    recurrence.fields.RecurrenceField(
                        blank=True, null=True, verbose_name="Recurrences"
                    ),
                ),
                (
                    "amends",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="amended_by",
                        to="core.calendarevent",
                        verbose_name="Amended base event",
                    ),
                ),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_%(app_label)s.%(class)s_set+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Calendar Event",
                "verbose_name_plural": "Calendar Events",
                "ordering": ["datetime_start", "date_start", "datetime_end", "date_end"],
            },
            bases=(aleksis.core.mixins.CalendarEventMixin, models.Model),
            managers=[
            ],
        ),
        migrations.CreateModel(
            name="Holiday",
            fields=[
                (
                    "calendarevent_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.calendarevent",
                    ),
                ),
                ("holiday_name", models.CharField(max_length=255, verbose_name="Name")),
            ],
            options={
                "verbose_name": "Holiday",
                "verbose_name_plural": "Holidays",
            },
            bases=("core.calendarevent",),
            managers=[
            ],
        ),
        migrations.AddConstraint(
            model_name="calendarevent",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("date_start__isnull", True), ("datetime_start__isnull", True), _negated=True
                ),
                name="datetime_start_or_date_start",
            ),
        ),
        migrations.AddConstraint(
            model_name="calendarevent",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("date_end__isnull", True), ("datetime_end__isnull", True), _negated=True
                ),
                name="datetime_end_or_date_end",
            ),
        ),
        migrations.AddConstraint(
            model_name="calendarevent",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("datetime_start__isnull", False), ("timezone__isnull", True), _negated=True
                ),
                name="timezone_if_datetime_start",
            ),
        ),
        migrations.AddConstraint(
            model_name="calendarevent",
            constraint=models.CheckConstraint(
                check=models.Q(
                    ("datetime_end__isnull", False), ("timezone__isnull", True), _negated=True
                ),
                name="timezone_if_datetime_end",
            ),
        ),
    ]



"""Management utilities for an AlekSIS installation."""

import os
import sys

import django.__main__
from django.core.management import execute_from_command_line


def aleksis_cmd():
    """Run django-admin command with correct settings path."""
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "aleksis.core.settings")
    sys.argv[0] = django.__main__.__file__
    execute_from_command_line(sys.argv)


if __name__ == "__main__":
    sys.exit(aleksis_cmd())



from datetime import datetime, time
from typing import Any, Callable, Dict, Sequence

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.core.exceptions import SuspiciousOperation, ValidationError
from django.db.models import QuerySet
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from allauth.account.adapter import get_adapter
from allauth.account.forms import SignupForm
from allauth.account.utils import setup_user_email
from dj_cleavejs import CleaveWidget
from django_select2.forms import ModelSelect2MultipleWidget, ModelSelect2Widget, Select2Widget
from dynamic_preferences.forms import PreferenceForm
from guardian.shortcuts import assign_perm
from invitations.forms import InviteForm
from maintenance_mode.core import get_maintenance_mode
from material import Fieldset, Layout, Row

from .mixins import ExtensibleForm, SchoolTermRelatedExtensibleForm
from .models import (
    Announcement,
    DashboardWidget,
    Group,
    GroupType,
    OAuthApplication,
    Person,
    PersonInvitation,
)
from .registries import (
    group_preferences_registry,
    person_preferences_registry,
    site_preferences_registry,
)
from .util.auth_helpers import AppScopes
from .util.core_helpers import get_site_preferences, queryset_rules_filter


class PersonForm(ExtensibleForm):
    """Form to edit or add a person object in the frontend."""

    layout = Layout(
        Fieldset(
            _("Base data"),
            "short_name",
            Row("user", "primary_group"),
            Row("first_name", "additional_name", "last_name"),
        ),
        Fieldset(_("Address"), Row("street", "housenumber"), Row("postal_code", "place")),
        Fieldset(_("Contact data"), "email", Row("phone_number", "mobile_number")),
        Fieldset(
            _("Advanced personal data"),
            Row("date_of_birth", "place_of_birth"),
            Row("sex"),
            Row("description"),
            Row("photo", "avatar"),
            "guardians",
        ),
    )

    class Meta:
        model = Person
        fields = [
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "street",
            "housenumber",
            "postal_code",
            "place",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "description",
            "photo",
            "avatar",
            "guardians",
            "primary_group",
        ]
        widgets = {
            "user": Select2Widget(attrs={"class": "browser-default"}),
            "primary_group": ModelSelect2Widget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "guardians": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }

    new_user = forms.CharField(
        required=False, label=_("New user"), help_text=_("Create a new account")
    )

    def __init__(self, *args, **kwargs):
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)

        if (
            request
            and self.instance
            and not request.user.has_perm("core.change_person", self.instance)
        ):
            # Disable non-editable fields
            allowed_person_fields = get_site_preferences()["account__editable_fields_person"]
            for field in self.fields:
                if field not in allowed_person_fields:
                    self.fields[field].disabled = True

    def clean(self) -> None:
        user = get_user_model()

        if self.cleaned_data.get("new_user", None):
            if self.cleaned_data.get("user", None):
                # The user selected both an existing user and provided a name to create a new one
                self.add_error(
                    "new_user",
                    _("You cannot set a new username when also selecting an existing user."),
                )
            elif user.objects.filter(username=self.cleaned_data["new_user"]).exists():
                # The user tried to create a new user with the name of an existing user
                self.add_error("new_user", _("This username is already in use."))
            else:
                # Create new User object and assign to form field for existing user
                new_user_obj = user.objects.create_user(
                    self.cleaned_data["new_user"],
                    self.instance.email,
                    first_name=self.instance.first_name,
                    last_name=self.instance.last_name,
                )

                self.cleaned_data["user"] = new_user_obj


class EditGroupForm(SchoolTermRelatedExtensibleForm):
    """Form to edit an existing group in the frontend."""

    layout = Layout(
        Fieldset(_("School term"), "school_term"),
        Fieldset(_("Common data"), "name", "short_name", "group_type"),
        Fieldset(_("Persons"), "members", "owners", "parent_groups"),
        Fieldset(_("Photo"), "photo", "avatar"),
    )

    class Meta:
        model = Group
        fields = [
            "school_term",
            "name",
            "short_name",
            "group_type",
            "members",
            "owners",
            "parent_groups",
            "photo",
            "avatar",
        ]
        widgets = {
            "members": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "owners": ModelSelect2MultipleWidget(
                search_fields=[
                    "first_name__icontains",
                    "last_name__icontains",
                    "short_name__icontains",
                ],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "parent_groups": ModelSelect2MultipleWidget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }


class AnnouncementForm(ExtensibleForm):
    """Form to create or edit an announcement in the frontend."""

    valid_from = forms.DateTimeField(required=False)
    valid_until = forms.DateTimeField(required=False)

    valid_from_date = forms.DateField(label=_("Date"))
    valid_from_time = forms.TimeField(label=_("Time"))

    valid_until_date = forms.DateField(label=_("Date"))
    valid_until_time = forms.TimeField(label=_("Time"))

    persons = forms.ModelMultipleChoiceField(
        queryset=Person.objects.all(),
        label=_("Persons"),
        required=False,
        widget=ModelSelect2MultipleWidget(
            search_fields=[
                "first_name__icontains",
                "last_name__icontains",
                "short_name__icontains",
            ],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
    )
    groups = forms.ModelMultipleChoiceField(
        queryset=None,
        label=_("Groups"),
        required=False,
        widget=ModelSelect2MultipleWidget(
            search_fields=[
                "name__icontains",
                "short_name__icontains",
            ],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
    )

    layout = Layout(
        Fieldset(
            _("From when until when should the announcement be displayed?"),
            Row("valid_from_date", "valid_from_time", "valid_until_date", "valid_until_time"),
        ),
        Fieldset(_("Who should see the announcement?"), Row("groups", "persons")),
        Fieldset(_("Write your announcement:"), "title", "description"),
    )

    def __init__(self, *args, **kwargs):
        if "instance" not in kwargs:
            # Default to today and whole day for new announcements
            kwargs["initial"] = {
                "valid_from_date": datetime.now(),
                "valid_from_time": time(0, 0),
                "valid_until_date": datetime.now(),
                "valid_until_time": time(23, 59),
            }
        else:
            announcement = kwargs["instance"]

            # Fill special fields from given announcement instance
            kwargs["initial"] = {
                "valid_from_date": announcement.valid_from.date(),
                "valid_from_time": announcement.valid_from.time(),
                "valid_until_date": announcement.valid_until.date(),
                "valid_until_time": announcement.valid_until.time(),
                "groups": announcement.get_recipients_for_model(Group),
                "persons": announcement.get_recipients_for_model(Person),
            }

        super().__init__(*args, **kwargs)

        self.fields["groups"].queryset = Group.objects.for_current_school_term_or_all()

    def clean(self):
        data = super().clean()

        # Combine date and time fields into datetime objects
        valid_from = datetime.combine(data["valid_from_date"], data["valid_from_time"])
        valid_until = datetime.combine(data["valid_until_date"], data["valid_until_time"])

        # Sanity check validity range
        if valid_until < datetime.now():
            raise ValidationError(
                _("You are not allowed to create announcements which are only valid in the past.")
            )
        elif valid_from > valid_until:
            raise ValidationError(
                _("The from date and time must be earlier then the until date and time.")
            )

        # Inject real time data if all went well
        data["valid_from"] = valid_from
        data["valid_until"] = valid_until

        # Ensure at least one group or one person is set as recipient
        if "groups" not in data and "persons" not in data:
            raise ValidationError(_("You need at least one recipient."))

        # Unwrap all recipients into single user objects and generate final list
        data["recipients"] = []
        data["recipients"] += data.get("groups", [])
        data["recipients"] += data.get("persons", [])

        return data

    def save(self, _=False):
        # Save announcement, respecting data injected in clean()
        if self.instance is None:
            self.instance = Announcement()
        self.instance.valid_from = self.cleaned_data["valid_from"]
        self.instance.valid_until = self.cleaned_data["valid_until"]
        self.instance.title = self.cleaned_data["title"]
        self.instance.description = self.cleaned_data["description"]
        self.instance.save()

        # Save recipients
        self.instance.recipients.all().delete()
        for recipient in self.cleaned_data["recipients"]:
            self.instance.recipients.create(recipient=recipient)
        self.instance.save()

        return self.instance

    class Meta:
        model = Announcement
        fields = [
            "valid_from_date",
            "valid_from_time",
            "valid_until_date",
            "valid_until_time",
            "groups",
            "persons",
            "title",
            "description",
        ]


class ChildGroupsForm(forms.Form):
    """Inline form for group editing to select child groups."""

    child_groups = forms.ModelMultipleChoiceField(queryset=Group.objects.all())


class SitePreferenceForm(PreferenceForm):
    """Form to edit site preferences."""

    registry = site_preferences_registry


class PersonPreferenceForm(PreferenceForm):
    """Form to edit preferences valid for one person."""

    registry = person_preferences_registry


class GroupPreferenceForm(PreferenceForm):
    """Form to edit preferences valid for members of a group."""

    registry = group_preferences_registry


class EditGroupTypeForm(forms.ModelForm):
    """Form to manage group types."""

    class Meta:
        model = GroupType
        fields = ["name", "description"]


class DashboardWidgetOrderForm(ExtensibleForm):
    pk = forms.ModelChoiceField(
        queryset=None,
        widget=forms.HiddenInput(attrs={"class": "pk-input"}),
    )
    order = forms.IntegerField(initial=0, widget=forms.HiddenInput(attrs={"class": "order-input"}))

    class Meta:
        model = DashboardWidget
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set queryset here to prevent problems with not migrated database due to special queryset
        self.fields["pk"].queryset = DashboardWidget.objects.all()


DashboardWidgetOrderFormSet = forms.formset_factory(
    form=DashboardWidgetOrderForm, max_num=0, extra=0
)


class InvitationCodeForm(forms.Form):
    """Form to enter an invitation code."""

    code = forms.CharField(
        label=_("Invitation code"),
        help_text=_("Please enter your invitation code."),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Calculate number of fields
        length = get_site_preferences()["auth__invite_code_length"]
        packet_size = get_site_preferences()["auth__invite_code_packet_size"]
        blocks = [
            packet_size,
        ] * length

        self.fields["code"].widget = CleaveWidget(blocks=blocks, delimiter="-", uppercase=True)


class PersonCreateInviteForm(InviteForm):
    """Custom form to create a person and invite them."""

    first_name = forms.CharField(label=_("First name"), required=True)
    last_name = forms.CharField(label=_("Last name"), required=True)

    layout = Layout(
        Row("first_name", "last_name"),
        Row("email"),
    )

    def clean_email(self):
        if Person.objects.filter(email=self.cleaned_data["email"]).exists():
            raise ValidationError(_("A person is using this e-mail address"))
        return super().clean_email()

    def save(self, email):
        person = Person.objects.create(
            first_name=self.cleaned_data["first_name"],
            last_name=self.cleaned_data["last_name"],
            email=email,
        )
        return PersonInvitation.create(email=email, person=person)


class SelectPermissionForm(forms.Form):
    """Select a permission to assign."""

    selected_permission = forms.ModelChoiceField(
        queryset=Permission.objects.all(),
        widget=ModelSelect2Widget(
            search_fields=["name__icontains", "codename__icontains"],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
    )


class AssignPermissionForm(forms.Form):
    """Assign a permission to user/groups for all/some objects."""

    layout = Layout(
        Fieldset(_("Who should get the permission?"), "groups", "persons"),
        Fieldset(_("On what?"), "objects", "all_objects"),
    )
    groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.all(),
        widget=ModelSelect2MultipleWidget(
            search_fields=["name__icontains", "short_name__icontains"],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        required=False,
    )
    persons = forms.ModelMultipleChoiceField(
        queryset=Person.objects.all(),
        widget=ModelSelect2MultipleWidget(
            search_fields=[
                "first_name__icontains",
                "last_name__icontains",
                "short_name__icontains",
            ],
            attrs={"data-minimum-input-length": 0, "class": "browser-default"},
        ),
        required=False,
    )

    objects = forms.ModelMultipleChoiceField(
        queryset=None,
        required=False,
        label=_("Select objects which the permission should be granted for:"),
    )
    all_objects = forms.BooleanField(
        required=False, label=_("Grant the permission for all objects")
    )

    def clean(self) -> Dict[str, Any]:
        """Clean form to ensure that at least one target and one type is selected."""
        cleaned_data = super().clean()
        if not cleaned_data.get("persons") and not cleaned_data.get("groups"):
            raise ValidationError(
                _("You must select at least one group or person which should get the permission.")
            )

        if not cleaned_data.get("objects") and not cleaned_data.get("all_objects"):
            raise ValidationError(
                _("You must grant the permission to all objects or to specific objects.")
            )
        return cleaned_data

    def __init__(self, *args, permission: Permission, **kwargs):
        self.permission = permission
        super().__init__(*args, **kwargs)

        model_class = self.permission.content_type.model_class()
        if model_class._meta.managed and not model_class._meta.abstract:
            queryset = model_class.objects.all()
        else:
            # The following queryset is just a dummy one. It has no real meaning.
            # We need it as there are permissions without real objects,
            # but we want to use the same form.
            queryset = Group.objects.none()
        self.fields["objects"].queryset = queryset
        search_fields = getattr(model_class, "get_filter_fields", lambda: [])()

        # Use select2 only if there are any searchable fields as it can't work without
        if search_fields:
            self.fields["objects"].widget = ModelSelect2MultipleWidget(
                search_fields=search_fields,
                queryset=queryset,
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            )

    def save_perms(self):
        """Save permissions for selected user/groups and selected/all objects."""
        persons = self.cleaned_data["persons"]
        groups = self.cleaned_data["groups"]
        all_objects = self.cleaned_data["all_objects"]
        objects = self.cleaned_data["objects"]
        permission_name = f"{self.permission.content_type.app_label}.{self.permission.codename}"

        # Create permissions for users
        for person in persons:
            if getattr(person, "user", None):
                # Global permission
                if all_objects:
                    assign_perm(permission_name, person.user)
                # Object permissions
                for instance in objects:
                    assign_perm(permission_name, person.user, instance)

        # Create permissions for users
        for group in groups:
            django_group = group.django_group
            # Global permission
            if all_objects:
                assign_perm(permission_name, django_group)
            # Object permissions
            for instance in objects:
                assign_perm(permission_name, django_group, instance)


class AccountRegisterForm(SignupForm, ExtensibleForm):
    """Form to register new user accounts."""

    class Meta:
        model = Person
        fields = [
            "first_name",
            "additional_name",
            "last_name",
            "street",
            "housenumber",
            "postal_code",
            "place",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "photo",
            "mobile_number",
            "phone_number",
            "short_name",
            "description",
        ]

    layout = Layout(
        Fieldset(
            _("Base data"),
            Row("first_name", "additional_name", "last_name"),
            "short_name",
        ),
        Fieldset(
            _("Address data"),
            Row("street", "housenumber"),
            Row("postal_code", "place"),
        ),
        Fieldset(_("Contact data"), Row("mobile_number", "phone_number")),
        Fieldset(
            _("Additional data"),
            Row("date_of_birth", "place_of_birth"),
            Row("sex", "photo"),
            "description",
        ),
        Fieldset(
            _("Account data"),
            "username",
            Row("email", "email2"),
            Row("password1", "password2"),
        ),
    )

    password1 = forms.CharField(label=_("Password"), widget=forms.PasswordInput)

    if settings.SIGNUP_PASSWORD_ENTER_TWICE:
        password2 = forms.CharField(label=_("Password (again)"), widget=forms.PasswordInput)

    def __init__(self, *args, **kwargs):
        request = kwargs.pop("request", None)
        super().__init__(*args, **kwargs)

        person = None
        if request.session.get("account_verified_email"):
            email = request.session["account_verified_email"]

            try:
                person = Person.objects.get(email=email)
            except (Person.DoesNotExist, Person.MultipleObjectsReturned) as exc:
                raise SuspiciousOperation from exc

        elif request.session.get("invitation_code"):
            try:
                invitation = PersonInvitation.objects.get(
                    key=request.session.get("invitation_code")
                )
            except PersonInvitation.DoesNotExist as exc:
                raise SuspiciousOperation from exc

            person = invitation.person

        if person:
            self.instance = person
            available_fields = [field.name for field in Person._meta.get_fields()]
            if person.email:
                self.fields["email"].disabled = True
                self.fields["email2"].disabled = True
                self.fields["email2"].initial = person.email
            for field in self.fields:
                if field in available_fields and getattr(person, field):
                    self.fields[field].disabled = True
                    self.fields[field].initial = getattr(person, field)

    def save(self, request):
        adapter = get_adapter(request)
        user = adapter.new_user(request)
        adapter.save_user(request, user, self)
        # Create person
        data = {}
        for field in Person._meta.get_fields():
            if field.name in self.cleaned_data:
                data[field.name] = self.cleaned_data[field.name]
        if self.instance:
            person_qs = Person.objects.filter(pk=self.instance.pk)
        else:
            person_qs = Person.objects.filter(email=data["email"])
            if not person_qs.exists() and get_site_preferences()["account__auto_create_person"]:
                Person.objects.create(user=user, **data)
        if person_qs.exists():
            person = person_qs.first()
            for field, value in data.items():
                setattr(person, field, value)
            person.user = user
            person.save()
        invitation_code = request.session.get("invitation_code")
        if invitation_code:
            from invitations.views import accept_invitation  # noqa

            try:
                invitation = PersonInvitation.objects.get(key=invitation_code)
            except PersonInvitation.DoesNotExist as exc:
                raise SuspiciousOperation from exc

            accept_invitation(invitation, request, user)
        self.custom_signup(request, user)
        setup_user_email(request, user, [])
        return user


class ActionForm(forms.Form):
    """Generic form for executing actions on multiple items of a queryset.

    This should be used together with a ``Table`` from django-tables2
    which includes a ``SelectColumn``.

    The queryset can be defined in two different ways:
    You can use ``get_queryset`` or provide ``queryset`` as keyword argument
    at the initialization of this form class.
    If both are declared, it will use the keyword argument.

    Any actions can be defined using the ``actions`` class attribute
    or overriding the method ``get_actions``.
    The actions use the same syntax like the Django Admin actions with one important difference:
    Instead of the related model admin,
    these actions will get the related ``ActionForm`` as first argument.
    Here you can see an example for such an action:

    .. code-block:: python

        from django.utils.translation import gettext as _

        def example_action(form, request, queryset):
            # Do something with this queryset

        example_action.short_description = _("Example action")

    If you can include the ``ActionForm`` like any other form in your views,
    but you must add the request as first argument.
    When the form is valid, you should run ``execute``:

    .. code-block:: python

        from aleksis.core.forms import ActionForm

        def your_view(request, ...):
            # Something
            action_form = ActionForm(request, request.POST or None, ...)
            if request.method == "POST" and form.is_valid():
                form.execute()

            # Something
    """

    layout = Layout("action")
    actions = []

    def get_actions(self) -> Sequence[Callable]:
        """Get all defined actions."""
        return self.actions

    def _get_actions_dict(self) -> dict[str, Callable]:
        """Get all defined actions as dictionary."""
        return {value.__name__: value for value in self.get_actions()}

    def _get_action_choices(self) -> list[tuple[str, str]]:
        """Get all defined actions as Django choices."""
        return [
            (value.__name__, getattr(value, "short_description", value.__name__))
            for value in self.get_actions()
        ]

    def get_queryset(self) -> QuerySet:
        """Get the related queryset."""
        raise NotImplementedError("Queryset necessary.")

    action = forms.ChoiceField(choices=[])
    selected_objects = forms.ModelMultipleChoiceField(queryset=None)

    def __init__(self, request: HttpRequest, *args, queryset: QuerySet = None, **kwargs):
        self.request = request
        self.queryset = queryset if isinstance(queryset, QuerySet) else self.get_queryset()
        super().__init__(*args, **kwargs)
        self.fields["selected_objects"].queryset = self.queryset
        self.fields["action"].choices = self._get_action_choices()

    def clean_action(self):
        action = self._get_actions_dict().get(self.cleaned_data["action"], None)
        if not action:
            raise ValidationError(_("The selected action does not exist."))
        return action

    def clean_selected_objects(self):
        action = self.cleaned_data["action"]
        if hasattr(action, "permission"):
            selected_objects = queryset_rules_filter(
                self.request, self.cleaned_data["selected_objects"], action.permission
            )
            if selected_objects.count() < self.cleaned_data["selected_objects"].count():
                raise ValidationError(
                    _("You do not have permission to run {} on all selected objects.").format(
                        getattr(action, "short_description", action.__name__)
                    )
                )
        return self.cleaned_data["selected_objects"]

    def execute(self) -> Any:
        """Execute the selected action on all selected objects.

        :return: the return value of the action
        """
        if self.is_valid():
            data = self.cleaned_data["selected_objects"]
            action = self.cleaned_data["action"]
            return action(None, self.request, data)

        raise TypeError("execute() must be called on a pre-validated form.")


class ListActionForm(ActionForm):
    """Generic form for executing actions on multiple items of a list of dictionaries.

    Sometimes you want to implement actions for data from different sources
    than querysets or even querysets from multiple models. For these cases,
    you can use this form.

    To provide an unique identification of each item, the dictionaries **must**
    include the attribute ``pk``. This attribute has to be unique for the whole list.
    If you don't mind this aspect, this will cause unexpected behavior.

    Any actions can be defined as described in ``ActionForm``, but, of course,
    the last argument won't be a queryset but a list of dictionaries.

    For further information on usage, you can take a look at ``ActionForm``.
    """

    selected_objects = forms.MultipleChoiceField(choices=[])

    def get_queryset(self):
        # Return None in order not to raise an unwanted exception
        return None

    def _get_dict(self) -> dict[str, dict]:
        """Get the items sorted by pk attribute."""
        return {item["pk"]: item for item in self.items}

    def _get_choices(self) -> list[tuple[str, str]]:
        """Get the items as Django choices."""
        return [(item["pk"], item["pk"]) for item in self.items]

    def _get_real_items(self, items: Sequence[dict]) -> list[dict]:
        """Get the real dictionaries from a list of pks."""
        items_dict = self._get_dict()
        real_items = []
        for item in items:
            if item not in items_dict:
                raise ValidationError(_("No valid selection."))
            real_items.append(items_dict[item])
        return real_items

    def clean_selected_objects(self) -> list[dict]:
        data = self.cleaned_data["selected_objects"]
        items = self._get_real_items(data)
        return items

    def __init__(self, request: HttpRequest, items, *args, **kwargs):
        self.items = items
        super().__init__(request, *args, **kwargs)
        self.fields["selected_objects"].choices = self._get_choices()


class OAuthApplicationForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["allowed_scopes"].widget = forms.SelectMultiple(
            choices=list(AppScopes().get_all_scopes().items())
        )

    class Meta:
        model = OAuthApplication
        fields = (
            "name",
            "icon",
            "client_id",
            "client_secret",
            "client_type",
            "algorithm",
            "allowed_scopes",
            "redirect_uris",
            "skip_authorization",
        )


class MaintenanceModeForm(forms.Form):
    maintenance_mode = forms.BooleanField(
        required=False,
        initial=not get_maintenance_mode(),
        widget=forms.HiddenInput,
    )



# flake8: noqa: DJ01
import base64
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, Any, Iterable, Iterator, List, Optional, Sequence, Union
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group as DjangoGroup
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator
from django.db import models
from django.db.models import Q, QuerySet
from django.dispatch import receiver
from django.forms.widgets import Media
from django.http import HttpRequest
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import date_format
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

import customidenticon
from cachalot.api import cachalot_disabled
from cache_memoize import cache_memoize
from calendarweek import CalendarWeek
from celery.result import AsyncResult
from celery_progress.backend import Progress
from ckeditor.fields import RichTextField
from django_celery_results.models import TaskResult
from django_ical.utils import build_rrule_from_recurrences_rrule, build_rrule_from_text
from dynamic_preferences.models import PerInstancePreferenceModel
from guardian.shortcuts import get_objects_for_user
from icalendar import vCalAddress, vText
from icalendar.prop import vRecur
from invitations import signals
from invitations.base_invitation import AbstractBaseInvitation
from invitations.models import Invitation
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel
from oauth2_provider.models import (
    AbstractAccessToken,
    AbstractApplication,
    AbstractGrant,
    AbstractIDToken,
    AbstractRefreshToken,
)
from phonenumber_field.modelfields import PhoneNumberField
from polymorphic.models import PolymorphicModel
from recurrence.fields import RecurrenceField
from timezone_field import TimeZoneField

from aleksis.core.data_checks import (
    DataCheck,
)

from .managers import (
    AlekSISBaseManagerWithoutMigrations,
    GroupManager,
    GroupQuerySet,
    HolidayQuerySet,
    InstalledWidgetsDashboardWidgetOrderManager,
    PolymorphicBaseManager,
    SchoolTermQuerySet,
    UninstallRenitentPolymorphicManager,
)
from .mixins import (
    CalendarEventMixin,
    ExtensibleModel,
    ExtensiblePolymorphicModel,
    GlobalPermissionModel,
    PureDjangoModel,
    RegistryObject,
    SchoolTermRelatedExtensibleModel,
)
from .tasks import send_notification
from .util.core_helpers import generate_random_code, get_site_preferences, now_tomorrow
from .util.email import send_email
from .util.model_helpers import ICONS

if TYPE_CHECKING:
    from django.contrib.auth.models import User

FIELD_CHOICES = (
    ("BooleanField", _("Boolean (Yes/No)")),
    ("CharField", _("Text (one line)")),
    ("DateField", _("Date")),
    ("DateTimeField", _("Date and time")),
    ("DecimalField", _("Decimal number")),
    ("EmailField", _("E-mail address")),
    ("IntegerField", _("Integer")),
    ("GenericIPAddressField", _("IP address")),
    ("NullBooleanField", _("Boolean or empty (Yes/No/Neither)")),
    ("TextField", _("Text (multi-line)")),
    ("TimeField", _("Time")),
    ("URLField", _("URL / Link")),
)


class SchoolTerm(ExtensibleModel):
    """School term model.

    This is used to manage start and end times of a school term and link data to it.
    """

    objects = AlekSISBaseManagerWithoutMigrations.from_queryset(SchoolTermQuerySet)()

    name = models.CharField(verbose_name=_("Name"), max_length=255)

    date_start = models.DateField(verbose_name=_("Start date"))
    date_end = models.DateField(verbose_name=_("End date"))

    @classmethod
    @cache_memoize(3600)
    def get_current(cls, day: Optional[date] = None):
        if not day:
            day = timezone.now().date()
        try:
            return cls.objects.on_day(day).first()
        except SchoolTerm.DoesNotExist:
            return None

    @classproperty
    def current(cls):
        return cls.get_current()

    def clean(self):
        """Ensure there is only one school term at each point of time."""
        if self.date_end < self.date_start:
            raise ValidationError(_("The start date must be earlier than the end date."))

        qs = SchoolTerm.objects.within_dates(self.date_start, self.date_end)
        if self.pk:
            qs = qs.exclude(pk=self.pk)
        if qs.exists():
            raise ValidationError(
                _("There is already a school term for this time or a part of this time.")
            )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("School term")
        verbose_name_plural = _("School terms")
        constraints = [
            models.UniqueConstraint(fields=["name"], name="unique_school_term_name"),
            models.UniqueConstraint(
                fields=["date_start", "date_end"],
                name="unique_school_term_dates",
            ),
        ]


class Person(ExtensibleModel):
    """Person model.

    A model describing any person related to a school, including, but not
    limited to, students, teachers and guardians (parents).
    """

    class Meta:
        ordering = ["last_name", "first_name"]
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")
        permissions = (
            ("view_address", _("Can view address")),
            ("view_contact_details", _("Can view contact details")),
            ("view_photo", _("Can view photo")),
            ("view_avatar", _("Can view avatar image")),
            ("view_person_groups", _("Can view persons groups")),
            ("view_personal_details", _("Can view personal details")),
        )
        constraints = [
            models.UniqueConstraint(
                fields=["short_name"], condition=~Q(short_name=""), name="unique_short_name"
            ),
            models.UniqueConstraint(fields=["email"], condition=~Q(email=""), name="unique_email"),
        ]

    icon_ = "account-outline"

    SEX_CHOICES = [("f", _("female")), ("m", _("male")), ("x", _("other"))]

    user = models.OneToOneField(
        get_user_model(),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="person",
        verbose_name=_("Linked user"),
    )

    first_name = models.CharField(verbose_name=_("First name"), max_length=255)
    last_name = models.CharField(verbose_name=_("Last name"), max_length=255)
    additional_name = models.CharField(
        verbose_name=_("Additional name(s)"), max_length=255, blank=True
    )

    short_name = models.CharField(
        verbose_name=_("Short name"),
        max_length=255,
        blank=True,
    )

    street = models.CharField(verbose_name=_("Street"), max_length=255, blank=True)
    housenumber = models.CharField(verbose_name=_("Street number"), max_length=255, blank=True)
    postal_code = models.CharField(verbose_name=_("Postal code"), max_length=255, blank=True)
    place = models.CharField(verbose_name=_("Place"), max_length=255, blank=True)

    phone_number = PhoneNumberField(verbose_name=_("Home phone"), blank=True)
    mobile_number = PhoneNumberField(verbose_name=_("Mobile phone"), blank=True)

    email = models.EmailField(verbose_name=_("E-mail address"), blank=True)

    date_of_birth = models.DateField(verbose_name=_("Date of birth"), blank=True, null=True)
    place_of_birth = models.CharField(verbose_name=_("Place of birth"), max_length=255, blank=True)
    sex = models.CharField(verbose_name=_("Sex"), max_length=1, choices=SEX_CHOICES, blank=True)

    photo = models.ImageField(
        verbose_name=_("Photo"),
        blank=True,
        null=True,
        help_text=_(
            "This is an official photo, used for official documents and for internal use cases."
        ),
    )

    avatar = models.ImageField(
        verbose_name=_("Display picture / Avatar"),
        blank=True,
        null=True,
        help_text=_("This is a picture or an avatar for public display."),
    )

    guardians = models.ManyToManyField(
        "self",
        verbose_name=_("Guardians / Parents"),
        symmetrical=False,
        related_name="children",
        blank=True,
    )

    primary_group = models.ForeignKey(
        "Group", models.SET_NULL, null=True, blank=True, verbose_name=_("Primary group")
    )

    description = models.TextField(verbose_name=_("Description"), blank=True)

    def get_absolute_url(self) -> str:
        return reverse("person_by_id", args=[self.id])

    @property
    def primary_group_short_name(self) -> Optional[str]:
        """Return the short_name field of the primary group related object."""
        if self.primary_group:
            return self.primary_group.short_name

    @primary_group_short_name.setter
    def primary_group_short_name(self, value: str) -> None:
        """
        Set the primary group related object by a short name.

        It uses the first existing group
        with this short name it can find, creating one
        if it can't find one.
        """
        group, created = Group.objects.get_or_create(short_name=value, defaults={"name": value})
        self.primary_group = group

    @property
    def full_name(self) -> str:
        """Full name of person in last name, first name order."""
        return f"{self.last_name}, {self.first_name}"

    @property
    def addressing_name(self) -> str:
        """Full name of person in format configured for addressing."""
        if self.preferences["notification__addressing_name_format"] == "last_first":
            return f"{self.last_name}, {self.first_name}"
        elif self.preferences["notification__addressing_name_format"] == "first_last":
            return f"{self.first_name} {self.last_name}"

    @property
    def mail_sender(self) -> str:
        """E-mail sender in "Name <email>" format."""
        return f'"{self.addressing_name}" <{self.email}>'

    @property
    def mail_sender_via(self) -> str:
        """E-mail sender for via addresses, in "Name via Site <email>" format."""
        site_mail = get_site_preferences()["mail__address"]
        site_name = get_site_preferences()["general__title"]

        return f'"{self.addressing_name} via {site_name}" <{site_mail}>'

    @property
    def age(self):
        """Age of the person at current time."""
        return self.age_at(timezone.now().date())

    def age_at(self, today):
        if self.date_of_birth:
            years = today.year - self.date_of_birth.year
            if self.date_of_birth.month > today.month or (
                self.date_of_birth.month == today.month and self.date_of_birth.day > today.day
            ):
                years -= 1
            return years

    @property
    def dashboard_widgets(self):
        return [
            w.widget
            for w in DashboardWidgetOrder.objects.filter(person=self, widget__active=True).order_by(
                "order"
            )
        ]

    @property
    def unread_notifications(self) -> QuerySet:
        """Get all unread notifications for this person."""
        return self.notifications.filter(read=False, send_at__lte=timezone.now())

    @property
    def unread_notifications_count(self) -> int:
        """Return the count of unread notifications for this person."""
        return self.unread_notifications.count()

    @property
    def initials(self):
        initials = ""
        if self.first_name:
            initials += self.first_name[0]
        if self.last_name:
            initials += self.last_name[0]
        return initials.upper() or "?"

    user_info_tracker = FieldTracker(fields=("first_name", "last_name", "email", "user_id"))

    @property
    @cache_memoize(60 * 60)
    def identicon_url(self):
        identicon = customidenticon.create(self.full_name, border=35)
        base64_data = base64.b64encode(identicon).decode("ascii")
        return f"data:image/png;base64,{base64_data}"

    @property
    def avatar_url(self):
        if self.avatar:
            return self.avatar.url
        else:
            return self.identicon_url

    def get_vcal_address(self, role: str = "REQ-PARTICIPANT") -> Optional[vCalAddress]:
        """Return a vCalAddress object for this person."""
        if not self.email:
            # vCalAddress requires an email address
            return None
        vcal = vCalAddress(f"MAILTO:{self.email}")
        vcal.params["cn"] = vText(self.full_name)
        vcal.params["ROLE"] = vText(role)
        return vcal

    def save(self, *args, **kwargs):
        # Determine all fields that were changed since last load
        changed = self.user_info_tracker.changed()

        super().save(*args, **kwargs)

        if self.pk is None or bool(changed):
            if "user_id" in changed:
                # Clear groups of previous Django user
                previous_user = changed["user_id"]
                if previous_user is not None:
                    get_user_model().objects.get(pk=previous_user).groups.clear()

            if self.user:
                if "first_name" in changed or "last_name" in changed or "email" in changed:
                    # Synchronise user fields to linked User object to keep it up to date
                    self.user.first_name = self.first_name
                    self.user.last_name = self.last_name
                    self.user.email = self.email
                    self.user.save()

                if "user_id" in changed:
                    # Synchronise groups to Django groups
                    for group in self.member_of.union(self.owner_of.all()).all():
                        group.save(force=True)

        # Select a primary group if none is set
        self.auto_select_primary_group()

    def __str__(self) -> str:
        return self.full_name

    @classmethod
    def maintain_default_data(cls):
        # Ensure we have an admin user
        user = get_user_model()
        if not user.objects.filter(is_superuser=True).exists():
            admin = user.objects.create_superuser(**settings.AUTH_INITIAL_SUPERUSER)
            admin.save()

    def auto_select_primary_group(
        self, pattern: Optional[str] = None, field: Optional[str] = None, force: bool = False
    ) -> None:
        """Auto-select the primary group among the groups the person is member of.

        Uses either the pattern passed as argument, or the pattern configured system-wide.

        Does not do anything if either no pattern is defined or the user already has
        a primary group, unless force is True.
        """
        pattern = pattern or get_site_preferences()["account__primary_group_pattern"]
        field = field or get_site_preferences()["account__primary_group_field"]

        if pattern and (force or not self.primary_group):
            self.primary_group = self.member_of.filter(**{f"{field}__regex": pattern}).first()

    def notify_about_changed_data(
        self, changed_fields: Iterable[str], recipients: Optional[List[str]] = None
    ):
        """Notify (configured) recipients about changed data of this person."""
        context = {"person": self, "changed_fields": changed_fields}
        recipients = recipients or [
            get_site_preferences()["account__person_change_notification_contact"]
        ]
        send_email(
            template_name="person_changed",
            from_email=self.mail_sender_via,
            headers={
                "Reply-To": self.mail_sender,
                "Sender": self.mail_sender,
            },
            recipient_list=recipients,
            context=context,
        )


class DummyPerson(Person):
    """A dummy person that is not stored into the database.

    Used to temporarily inject a Person object into a User.
    """

    class Meta:
        managed = False
        proxy = True

    is_dummy = True

    def save(self, *args, **kwargs):
        # Do nothing, not even call Model's save(), so this is never persisted
        pass


class Group(SchoolTermRelatedExtensibleModel):
    """Group model.

    Any kind of group of persons in a school, including, but not limited
    classes, clubs, and the like.
    """

    objects = GroupManager.from_queryset(GroupQuerySet)()

    class Meta:
        ordering = ["short_name", "name"]
        verbose_name = _("Group")
        verbose_name_plural = _("Groups")
        permissions = (
            ("assign_child_groups_to_groups", _("Can assign child groups to groups")),
            ("view_group_stats", _("Can view statistics about group.")),
        )
        constraints = [
            models.UniqueConstraint(
                fields=["school_term", "name"], name="unique_school_term_name_group"
            ),
            models.UniqueConstraint(
                fields=["school_term", "short_name"], name="unique_school_term_short_name_group"
            ),
        ]

    icon_ = "account-multiple-outline"

    name = models.CharField(verbose_name=_("Long name"), max_length=255)
    short_name = models.CharField(  # noqa
        verbose_name=_("Short name"),
        max_length=255,
        blank=True,
        null=True,
    )

    members = models.ManyToManyField(
        "Person",
        related_name="member_of",
        blank=True,
        through="PersonGroupThrough",
        verbose_name=_("Members"),
    )
    owners = models.ManyToManyField(
        "Person", related_name="owner_of", blank=True, verbose_name=_("Owners")
    )

    parent_groups = models.ManyToManyField(
        "self",
        related_name="child_groups",
        symmetrical=False,
        verbose_name=_("Parent groups"),
        blank=True,
    )

    group_type = models.ForeignKey(
        "GroupType",
        on_delete=models.SET_NULL,
        related_name="type",
        verbose_name=_("Type of group"),
        null=True,
        blank=True,
    )

    photo = models.ImageField(
        verbose_name=_("Photo"),
        blank=True,
        null=True,
        help_text=_(
            "This is an official photo, used for official documents and for internal use cases."
        ),
    )
    avatar = models.ImageField(
        verbose_name=_("Display picture / Avatar"),
        blank=True,
        null=True,
        help_text=_("This is a picture or an avatar for public display."),
    )

    def get_absolute_url(self) -> str:
        return reverse("group_by_id", args=[self.id])

    @property
    def announcement_recipients(self):
        """Flat list of all members and owners to fulfill announcement API contract."""
        return list(self.members.all()) + list(self.owners.all())

    @property
    def get_group_stats(self) -> dict:
        """Get stats about a given group"""
        stats = {}

        stats["members"] = len(self.members.all())

        ages = [person.age for person in self.members.filter(date_of_birth__isnull=False)]

        if ages:
            stats["age_avg"] = sum(ages) / len(ages)
            stats["age_range_min"] = min(ages)
            stats["age_range_max"] = max(ages)

        return stats

    def __str__(self) -> str:
        if self.school_term:
            return f"{self.name} ({self.short_name}) ({self.school_term})"
        else:
            return f"{self.name} ({self.short_name})"

    group_info_tracker = FieldTracker(fields=("name", "short_name"))

    def save(self, force: bool = False, *args, **kwargs):
        # Determine state of object in relation to database
        dirty = self.pk is None or bool(self.group_info_tracker.changed())

        super().save(*args, **kwargs)

        if force or dirty:
            # Synchronise group to Django group with same name
            dj_group, _ = DjangoGroup.objects.get_or_create(name=self.name)
            dj_group.user_set.set(
                list(
                    self.members.filter(user__isnull=False)
                    .values_list("user", flat=True)
                    .union(self.owners.filter(user__isnull=False).values_list("user", flat=True))
                )
            )
            dj_group.save()

    @property
    def django_group(self):
        """Get Django group for this group."""
        dj_group, _ = DjangoGroup.objects.get_or_create(name=self.name)
        return dj_group


class PersonGroupThrough(ExtensibleModel):
    """Through table for many-to-many relationship of group members.

    It does not have any fields on its own; these are generated upon instantiation
    by inspecting the additional fields selected for the linked group.
    """

    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    person = models.ForeignKey(Person, on_delete=models.CASCADE)


@receiver(models.signals.m2m_changed, sender=PersonGroupThrough)
@receiver(models.signals.m2m_changed, sender=Group.owners.through)
def save_group_on_m2m_changed(
    sender: Union[PersonGroupThrough, Group.owners.through],
    instance: models.Model,
    action: str,
    reverse: bool,
    model: models.Model,
    pk_set: Optional[set],
    **kwargs,
) -> None:
    """Ensure user and group data is synced to Django's models.

    AlekSIS maintains personal information and group meta-data / membership
    in its Person and Group models. As third-party libraries have no knowledge
    about this, we need to keep django.contrib.auth in sync.

    This signal handler triggers a save of group objects whenever a membership
    changes. The save() code will decide whether to update the Django objects
    or not.
    """
    if action not in ("post_add", "post_remove", "post_clear"):
        # Only trigger once, after the change was applied to the database
        return

    if reverse:
        # Relationship was changed on the Person side, saving all groups
        # that have been touched there
        for group in model.objects.filter(pk__in=pk_set):
            group.save(force=True)
    else:
        # Relationship was changed on the Group side
        instance.save(force=True)


class Activity(ExtensibleModel, TimeStampedModel):
    """Activity of a user to trace some actions done in AlekSIS in displayable form."""

    user = models.ForeignKey(
        "Person", on_delete=models.CASCADE, related_name="activities", verbose_name=_("User")
    )

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"))

    app = models.CharField(max_length=100, verbose_name=_("Application"))

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Activity")
        verbose_name_plural = _("Activities")


class Notification(ExtensibleModel, TimeStampedModel):
    """Notification to submit to a user."""

    sender = models.CharField(max_length=100, verbose_name=_("Sender"))
    recipient = models.ForeignKey(
        "Person",
        on_delete=models.CASCADE,
        related_name="notifications",
        verbose_name=_("Recipient"),
    )

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"))
    link = models.URLField(blank=True, verbose_name=_("Link"))

    icon = models.CharField(
        max_length=50, choices=ICONS, verbose_name=_("Icon"), default="information-outline"
    )

    send_at = models.DateTimeField(default=timezone.now, verbose_name=_("Send notification at"))

    read = models.BooleanField(default=False, verbose_name=_("Read"))
    sent = models.BooleanField(default=False, verbose_name=_("Sent"))

    def __str__(self):
        return str(self.title)

    def save(self, **kwargs):
        super().save(**kwargs)
        if not self.sent and self.send_at <= timezone.now():
            self.send()
            super().save(**kwargs)

    def send(self, resend: bool = False) -> Optional[AsyncResult]:
        """Send the notification to the recipient."""
        if not self.sent or resend:
            return send_notification.delay(self.pk, resend=resend)

    class Meta:
        verbose_name = _("Notification")
        verbose_name_plural = _("Notifications")


class AnnouncementQuerySet(models.QuerySet):
    """Queryset for announcements providing time-based utility functions."""

    def relevant_for(self, obj: Union[models.Model, models.QuerySet]) -> models.QuerySet:
        """Get all relevant announcements.

        Get a QuerySet with all announcements relevant for a certain Model (e.g. a Group)
        or a set of models in a QuerySet.
        """
        if isinstance(obj, models.QuerySet):
            ct = ContentType.objects.get_for_model(obj.model)
            pks = list(obj.values_list("pk", flat=True))
        else:
            ct = ContentType.objects.get_for_model(obj)
            pks = [obj.pk]

        return self.filter(recipients__content_type=ct, recipients__recipient_id__in=pks)

    def at_time(self, when: Optional[datetime] = None) -> models.QuerySet:
        """Get all announcements at a certain time."""
        when = when or timezone.now()

        # Get announcements by time
        announcements = self.filter(valid_from__lte=when, valid_until__gte=when)

        return announcements

    def on_date(self, when: Optional[date] = None) -> models.QuerySet:
        """Get all announcements at a certain date."""
        when = when or timezone.now().date()

        # Get announcements by time
        announcements = self.filter(valid_from__date__lte=when, valid_until__date__gte=when)

        return announcements

    def within_days(self, start: date, stop: date) -> models.QuerySet:
        """Get all announcements valid for a set of days."""
        # Get announcements
        announcements = self.filter(valid_from__date__lte=stop, valid_until__date__gte=start)

        return announcements

    def for_person(self, person: Person) -> List:
        """Get all announcements for one person."""
        # Filter by person
        announcements_for_person = []
        for announcement in self:
            if person in announcement.recipient_persons:
                announcements_for_person.append(announcement)

        return announcements_for_person


class Announcement(ExtensibleModel):
    """Announcement model.

    Persistent announcement to display to groups or persons in various places during a
    specific time range.
    """

    objects = models.Manager.from_queryset(AnnouncementQuerySet)()

    title = models.CharField(max_length=150, verbose_name=_("Title"))
    description = models.TextField(max_length=500, verbose_name=_("Description"), blank=True)
    link = models.URLField(blank=True, verbose_name=_("Link to detailed view"))

    valid_from = models.DateTimeField(
        verbose_name=_("Date and time from when to show"), default=timezone.now
    )
    valid_until = models.DateTimeField(
        verbose_name=_("Date and time until when to show"),
        default=now_tomorrow,
    )

    @property
    def recipient_persons(self) -> Sequence[Person]:
        """Return a list of Persons this announcement is relevant for."""
        persons = []
        for recipient in self.recipients.all():
            persons += recipient.persons
        return persons

    def get_recipients_for_model(self, obj: Union[models.Model]) -> Sequence[models.Model]:
        """Get all recipients.

        Get all recipients for this announcement
        with a special content type (provided through model)
        """
        ct = ContentType.objects.get_for_model(obj)
        return [r.recipient for r in self.recipients.filter(content_type=ct)]

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Announcement")
        verbose_name_plural = _("Announcements")


class AnnouncementRecipient(ExtensibleModel):
    """Announcement recipient model.

    Generalisation of a recipient for an announcement, used to wrap arbitrary
    objects that can receive announcements.

    Contract: Objects to serve as recipient have a property announcement_recipients
    returning a flat list of Person objects.
    """

    announcement = models.ForeignKey(
        Announcement, on_delete=models.CASCADE, related_name="recipients"
    )

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    recipient_id = models.PositiveIntegerField()
    recipient = GenericForeignKey("content_type", "recipient_id")

    @property
    def persons(self) -> Sequence[Person]:
        """Return a list of Persons selected by this recipient object.

        If the recipient is a Person, return that object. If not, it returns the list
        from the announcement_recipients field on the target model.
        """
        if isinstance(self.recipient, Person):
            return [self.recipient]
        else:
            return getattr(self.recipient, "announcement_recipients", [])

    def __str__(self):
        return str(self.recipient)

    class Meta:
        verbose_name = _("Announcement recipient")
        verbose_name_plural = _("Announcement recipients")


class DashboardWidget(RegistryObject, PolymorphicModel, PureDjangoModel):
    """Base class for dashboard widgets on the index page."""

    objects = UninstallRenitentPolymorphicManager()

    @staticmethod
    def get_media(widgets: Union[QuerySet, Iterable]):
        """Return all media required to render the selected widgets."""
        media = Media()
        for widget in widgets:
            media = media + widget.media
        return media

    template = None
    template_broken = "core/dashboard_widget/dashboardwidget_broken.html"
    media = Media()

    title = models.CharField(max_length=150, verbose_name=_("Widget Title"))
    active = models.BooleanField(verbose_name=_("Activate Widget"))
    broken = models.BooleanField(verbose_name=_("Widget is broken"), default=False)

    size_s = models.PositiveSmallIntegerField(
        verbose_name=_("Size on mobile devices"),
        help_text=_("<= 600 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=12,
    )
    size_m = models.PositiveSmallIntegerField(
        verbose_name=_("Size on tablet devices"),
        help_text=_("> 600 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=12,
    )
    size_l = models.PositiveSmallIntegerField(
        verbose_name=_("Size on desktop devices"),
        help_text=_("> 992 px, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=6,
    )
    size_xl = models.PositiveSmallIntegerField(
        verbose_name=_("Size on large desktop devices"),
        help_text=_("> 1200 px>, 12 columns"),
        validators=[MaxValueValidator(12)],
        default=4,
    )

    def _get_context_safe(self, request):
        if self.broken:
            return {"title": self.title}
        return self.get_context(request)

    def get_context(self, request):
        """Get the context dictionary to pass to the widget template."""
        raise NotImplementedError("A widget subclass needs to implement the get_context method.")

    def get_template(self):
        """Get template.

        Get the template to render the widget with. Defaults to the template attribute,
        but can be overridden to allow more complex template generation scenarios. If
        the widget is marked as broken, the template_broken attribute will be returned.
        """
        if self.broken:
            return self.template_broken
        if not self.template:
            raise NotImplementedError("A widget subclass needs to define a template.")
        return self.template

    def __str__(self):
        return self.title

    class Meta:
        permissions = (("edit_default_dashboard", _("Can edit default dashboard")),)
        verbose_name = _("Dashboard Widget")
        verbose_name_plural = _("Dashboard Widgets")


class ExternalLinkWidget(DashboardWidget):
    template = "core/dashboard_widget/external_link_widget.html"

    url = models.URLField(verbose_name=_("URL"))
    icon_url = models.URLField(verbose_name=_("Icon URL"))

    def get_context(self, request):
        return {"title": self.title, "url": self.url, "icon_url": self.icon_url}

    class Meta:
        verbose_name = _("External link widget")
        verbose_name_plural = _("External link widgets")


class StaticContentWidget(DashboardWidget):
    template = "core/dashboard_widget/static_content_widget.html"

    content = RichTextField(verbose_name=_("Content"))

    def get_context(self, request):
        return {"title": self.title, "content": self.content}

    class Meta:
        verbose_name = _("Static content widget")
        verbose_name_plural = _("Static content widgets")


class DashboardWidgetOrder(ExtensibleModel):
    widget = models.ForeignKey(
        DashboardWidget, on_delete=models.CASCADE, verbose_name=_("Dashboard widget")
    )
    person = models.ForeignKey(
        Person, on_delete=models.CASCADE, verbose_name=_("Person"), null=True, blank=True
    )
    order = models.PositiveIntegerField(verbose_name=_("Order"))
    default = models.BooleanField(default=False, verbose_name=_("Part of the default dashboard"))

    objects = InstalledWidgetsDashboardWidgetOrderManager()

    @classproperty
    def default_dashboard_widgets(cls):
        """Get default order for dashboard widgets."""
        return [
            w.widget
            for w in cls.objects.filter(person=None, default=True, widget__active=True).order_by(
                "order"
            )
        ]

    class Meta:
        verbose_name = _("Dashboard widget order")
        verbose_name_plural = _("Dashboard widget orders")


class CustomMenu(ExtensibleModel):
    """A custom menu to display in the footer."""

    name = models.CharField(max_length=100, verbose_name=_("Menu ID"))

    def __str__(self):
        return self.name if self.name != "" else self.id

    @classmethod
    @cache_memoize(3600)
    def get_default(cls, name):
        """Get a menu by name or create if it does not exist."""
        menu, _ = cls.objects.prefetch_related("items").get_or_create(name=name)
        return menu

    class Meta:
        verbose_name = _("Custom menu")
        verbose_name_plural = _("Custom menus")
        constraints = [
            models.UniqueConstraint(fields=["name"], name="unique_menu_name"),
        ]


class CustomMenuItem(ExtensibleModel):
    """Single item in a custom menu."""

    menu = models.ForeignKey(
        CustomMenu, models.CASCADE, verbose_name=_("Menu"), related_name="items"
    )
    name = models.CharField(max_length=150, verbose_name=_("Name"))
    url = models.URLField(verbose_name=_("Link"))
    icon = models.CharField(max_length=50, blank=True, choices=ICONS, verbose_name=_("Icon"))

    def __str__(self):
        return f"[{self.menu}] {self.name}"

    class Meta:
        verbose_name = _("Custom menu item")
        verbose_name_plural = _("Custom menu items")
        constraints = [
            models.UniqueConstraint(fields=["menu", "name"], name="unique_name_per_menu"),
        ]

    def get_absolute_url(self):
        return reverse("admin:core_custommenuitem_change", args=[self.id])


class DynamicRoute(RegistryObject):
    """Define a dynamic route.

    Dynamic routes should be used to register Vue routes dynamically, e. g.
    when an app is supposed to show menu items for dynamically creatable objects.
    """


class GroupType(ExtensibleModel):
    """Group type model.

    Descriptive type of a group; used to tag groups and for apps to distinguish
    how to display or handle a certain group.
    """

    name = models.CharField(verbose_name=_("Title of type"), max_length=50)
    description = models.CharField(verbose_name=_("Description"), max_length=500)

    def __str__(self) -> str:
        return self.name

    class Meta:
        verbose_name = _("Group type")
        verbose_name_plural = _("Group types")
        constraints = [
            models.UniqueConstraint(fields=["name"], name="unique_group_type_name"),
        ]


class GlobalPermissions(GlobalPermissionModel):
    """Container for global permissions."""

    class Meta(GlobalPermissionModel.Meta):
        permissions = (
            ("view_system_status", _("Can view system status")),
            ("manage_data", _("Can manage data")),
            ("impersonate", _("Can impersonate")),
            ("search", _("Can use search")),
            ("change_site_preferences", _("Can change site preferences")),
            ("change_person_preferences", _("Can change person preferences")),
            ("change_group_preferences", _("Can change group preferences")),
            ("test_pdf", _("Can test PDF generation")),
            ("invite", _("Can invite persons")),
            ("view_birthday_calendar", _("Can view birthday calendar")),
        )


class PersonPreferenceModel(PerInstancePreferenceModel, PureDjangoModel):
    """Preference model to hold pereferences valid for a person."""

    instance = models.ForeignKey(Person, on_delete=models.CASCADE)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = "core"


class GroupPreferenceModel(PerInstancePreferenceModel, PureDjangoModel):
    """Preference model to hold pereferences valid for members of a group."""

    instance = models.ForeignKey(Group, on_delete=models.CASCADE)

    class Meta(PerInstancePreferenceModel.Meta):
        app_label = "core"


class DataCheckResult(ExtensibleModel):
    """Save the result of a data check for a specific object."""

    data_check = models.CharField(
        max_length=255,
        verbose_name=_("Related data check task"),
        choices=DataCheck.data_checks_choices,
    )

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=255)
    related_object = GenericForeignKey("content_type", "object_id")

    solved = models.BooleanField(default=False, verbose_name=_("Issue solved"))
    sent = models.BooleanField(default=False, verbose_name=_("Notification sent"))

    @property
    def related_check(self) -> DataCheck:
        return DataCheck.registered_objects_dict[self.data_check]

    def solve(self, solve_option: str = "default"):
        self.related_check.solve(self, solve_option)

    def __str__(self):
        return f"{self.related_object}: {self.related_check.problem_name}"

    class Meta:
        verbose_name = _("Data check result")
        verbose_name_plural = _("Data check results")
        permissions = (
            ("run_data_checks", _("Can run data checks")),
            ("solve_data_problem", _("Can solve data check problems")),
        )


class PersonInvitation(AbstractBaseInvitation, PureDjangoModel):
    """Custom model for invitations to allow to generate invitations codes without email address."""

    email = models.EmailField(verbose_name=_("E-Mail address"), blank=True)
    person = models.ForeignKey(
        Person, on_delete=models.CASCADE, blank=True, related_name="invitation", null=True
    )

    def __str__(self) -> str:
        return f"{self.email} ({self.inviter})"

    @classmethod
    def create(cls, email, inviter=None, **kwargs):
        length = get_site_preferences()["auth__invite_code_length"]
        packet_size = get_site_preferences()["auth__invite_code_packet_size"]
        code = generate_random_code(length, packet_size)

        instance = cls.objects.create(email=email, inviter=inviter, key=code, **kwargs)
        return instance

    def send_invitation(self, request, **kwargs):
        """Send the invitation email to the person."""
        invite_url = reverse("invitations:accept-invite", args=[self.key])
        invite_url = request.build_absolute_uri(invite_url).replace("/django", "")
        context = kwargs
        context.update(
            {
                "invite_url": invite_url,
                "site_name": get_site_preferences()["general__title"],
                "email": self.email,
                "inviter": self.inviter,
                "person": self.person,
            },
        )

        send_email(template_name="invitation", recipient_list=[self.email], context=context)

        self.sent = timezone.now()
        self.save()

        signals.invite_url_sent.send(
            sender=self.__class__,
            instance=self,
            invite_url_sent=invite_url,
            inviter=self.inviter,
        )

    key_expired = Invitation.key_expired
    send_invitation = send_invitation


class PDFFile(ExtensibleModel):
    """Link to a rendered PDF file."""

    def _get_default_expiration():  # noqa
        return timezone.now() + timedelta(minutes=get_site_preferences()["general__pdf_expiration"])

    person = models.ForeignKey(
        to=Person,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("Owner"),
        related_name="pdf_files",
    )
    expires_at = models.DateTimeField(
        verbose_name=_("File expires at"), default=_get_default_expiration
    )
    html_file = models.FileField(
        upload_to="pdfs/", verbose_name=_("Generated HTML file"), blank=True, null=True
    )
    file = models.FileField(
        upload_to="pdfs/", blank=True, null=True, verbose_name=_("Generated PDF file")
    )

    def __str__(self):
        return f"{self.person} ({self.pk})"

    class Meta:
        verbose_name = _("PDF file")
        verbose_name_plural = _("PDF files")


class TaskUserAssignment(ExtensibleModel):
    task_result = models.ForeignKey(
        TaskResult, on_delete=models.CASCADE, verbose_name=_("Task result")
    )
    user = models.ForeignKey(
        get_user_model(), on_delete=models.CASCADE, verbose_name=_("Task user")
    )

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    back_url = models.URLField(verbose_name=_("Back URL"), blank=True)
    progress_title = models.CharField(max_length=255, verbose_name=_("Progress title"), blank=True)
    error_message = models.TextField(verbose_name=_("Error message"), blank=True)
    success_message = models.TextField(verbose_name=_("Success message"), blank=True)
    redirect_on_success_url = models.URLField(verbose_name=_("Redirect on success URL"), blank=True)
    additional_button_title = models.CharField(
        max_length=255, verbose_name=_("Additional button title"), blank=True
    )
    additional_button_url = models.URLField(verbose_name=_("Additional button URL"), blank=True)
    additional_button_icon = models.CharField(
        max_length=255, verbose_name=_("Additional button icon"), blank=True
    )
    result_fetched = models.BooleanField(default=False, verbose_name=_("Result fetched"))

    @classmethod
    def create_for_task_id(cls, task_id: str, user: "User") -> "TaskUserAssignment":
        # Use get_or_create to ensure the TaskResult exists
        # django-celery-results will later add the missing information
        with cachalot_disabled():
            result, __ = TaskResult.objects.get_or_create(task_id=task_id)
        return cls.objects.create(task_result=result, user=user)

    def get_progress(self) -> dict[str, any]:
        """Get progress information for this task."""
        progress = Progress(AsyncResult(self.task_result.task_id))
        return progress.get_info()

    def get_progress_with_meta(self) -> dict[str, any]:
        """Get progress information for this task."""
        progress = self.get_progress()
        progress["meta"] = self
        return progress

    def create_notification(self) -> Optional[Notification]:
        """Create a notification for this task."""
        progress = self.get_progress()
        if progress["state"] == "SUCCESS":
            title = _("Background task completed successfully")
            description = _("The background task '{}' has been completed successfully.").format(
                self.title
            )
            icon = "check-circle-outline"

        elif progress["state"] == "FAILURE":
            title = _("Background task failed")
            description = _("The background task '{}' has failed.").format(self.title)
            icon = "alert-octagon-outline"
        else:
            # Task not yet finished
            return

        link = urljoin(settings.BASE_URL, self.get_absolute_url())

        notification = Notification(
            sender=_("Background task"),
            recipient=self.user.person,
            title=title,
            description=description,
            link=link,
            icon=icon,
        )
        notification.save()
        return notification

    def get_absolute_url(self) -> str:
        return f"/celery_progress/{self.task_result.task_id}"

    class Meta:
        verbose_name = _("Task user assignment")
        verbose_name_plural = _("Task user assignments")


class UserAdditionalAttributes(models.Model, PureDjangoModel):
    """Additional attributes for Django user accounts.

    These attributes are explicitly linked to a User, not to a Person.
    """

    user = models.OneToOneField(
        get_user_model(),
        on_delete=models.CASCADE,
        related_name="additional_attributes",
        verbose_name=_("Linked user"),
    )

    attributes = models.JSONField(verbose_name=_("Additional attributes"), default=dict)

    def __str__(self):
        return str(self.user)

    @classmethod
    def get_user_attribute(
        cls, username: str, attribute: str, default: Optional[Any] = None
    ) -> Any:
        """Get a user attribute for a user by name."""
        try:
            attributes = cls.objects.get(user__username=username)
        except cls.DoesNotExist:
            return default

        return attributes.attributes.get(attribute, default)

    @classmethod
    def set_user_attribute(cls, username: str, attribute: str, value: Any):
        """Set a user attribute for a user by name.

        Raises DoesNotExist if a username for a non-existing Django user is passed.
        """
        user = get_user_model().objects.get(username=username)
        attributes, __ = cls.objects.update_or_create(user=user)

        attributes.attributes[attribute] = value
        attributes.save()


class OAuthApplication(AbstractApplication):
    """Modified OAuth application class that supports Grant Flows configured in preferences."""

    # Override grant types to make field optional
    authorization_grant_type = models.CharField(
        max_length=32, choices=AbstractApplication.GRANT_TYPES, blank=True
    )

    # Optional list of alloewd scopes
    allowed_scopes = ArrayField(
        models.CharField(max_length=255),
        verbose_name=_("Allowed scopes that clients can request"),
        null=True,
        blank=True,
    )

    icon = models.ImageField(
        verbose_name=_("Icon"),
        blank=True,
        null=True,
        help_text=_(
            "This image will be shown as icon in the authorization flow. It should be squared."
        ),
    )

    def allows_grant_type(self, *grant_types: set[str]) -> bool:
        allowed_grants = get_site_preferences()["auth__oauth_allowed_grants"]

        return bool(set(allowed_grants) & set(grant_types))

    def get_absolute_url(self):
        return reverse("oauth2_application", args=[self.id])


class OAuthGrant(AbstractGrant):
    """Placeholder for customising the Grant model."""

    pass


class OAuthAccessToken(AbstractAccessToken):
    """Placeholder for customising the AccessToken model."""

    pass


class OAuthIDToken(AbstractIDToken):
    """Placeholder for customising the IDToken model."""

    pass


class OAuthRefreshToken(AbstractRefreshToken):
    """Placeholder for customising the RefreshToken model."""

    pass


class Room(ExtensibleModel):
    short_name = models.CharField(verbose_name=_("Short name"), max_length=255)
    name = models.CharField(verbose_name=_("Long name"), max_length=255)

    icon_ = "door"

    def __str__(self) -> str:
        return f"{self.name} ({self.short_name})"

    def get_absolute_url(self) -> str:
        return reverse("timetable", args=["room", self.id])

    class Meta:
        permissions = (("view_room_timetable", _("Can view room timetable")),)
        ordering = ["name", "short_name"]
        verbose_name = _("Room")
        verbose_name_plural = _("Rooms")
        constraints = [
            models.UniqueConstraint(fields=["short_name"], name="unique_room_short_name"),
        ]


class CalendarEvent(CalendarEventMixin, ExtensiblePolymorphicModel):
    """A planned event in a calendar.

    To make use of this model, you need to inherit from this model.
    Every subclass of this model represents a certain calendar (feed).
    It therefore needs to set the basic attributes of the calendar like
    described in the documentation of `CalendarEventMixin`.

    Furthermore, every `value_*` method from `CalendarEventMixin`
    can be implemented to provide additional data (either static or dynamic).
    Some like start and end date are pre-implemented in this model. Others, like
    `value_title` need to be implemented in the subclass. Some methods are
    also optional, like `value_location` or `value_description`.
    Please refer to the documentation of `CalendarEventMixin` for more information.
    """

    datetime_start = models.DateTimeField(
        verbose_name=_("Start date and time"), null=True, blank=True
    )
    datetime_end = models.DateTimeField(verbose_name=_("End date and time"), null=True, blank=True)
    timezone = TimeZoneField(verbose_name=_("Timezone"), null=True, blank=True)
    date_start = models.DateField(verbose_name=_("Start date"), null=True, blank=True)
    date_end = models.DateField(verbose_name=_("End date"), null=True, blank=True)
    recurrences = RecurrenceField(verbose_name=_("Recurrences"), null=True, blank=True)
    amends = models.ForeignKey(
        "self",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        verbose_name=_("Amended base event"),
        related_name="amended_by",
    )

    def provide_list_in_timezone(self, seq):
        """Provide a list of datetimes in the saved timezone."""
        return [dt.astimezone(self.timezone) if isinstance(dt, datetime) else dt for dt in seq]

    @classmethod
    def value_title(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> str:
        """Return the title of the calendar event."""
        raise NotImplementedError()

    @classmethod
    def value_start_datetime(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Union[datetime, date]:
        """Return the start datetime of the calendar event."""
        if reference_object.datetime_start:
            return reference_object.datetime_start.astimezone(reference_object.timezone)
        return reference_object.date_start

    @classmethod
    def value_end_datetime(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Union[datetime, date, None]:
        """Return the end datetime of the calendar event."""
        if reference_object.datetime_end:
            return reference_object.datetime_end.astimezone(reference_object.timezone)
        if reference_object.date_end == reference_object.date_start:
            # Rule for all day events: If the event is only one day long,
            # the end date has to be empty
            return None
        return reference_object.date_end

    @classmethod
    def value_rrule(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Optional[vRecur]:
        """Return the rrule of the calendar event."""
        if not reference_object.recurrences or not reference_object.recurrences.rrules:
            return None
        # iCal only supports one RRULE per event as per RFC 5545 (change to older RFC 2445)
        return build_rrule_from_recurrences_rrule(reference_object.recurrences.rrules[0])

    @classmethod
    def value_rdate(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Optional[list[datetime]]:
        """Return the rdate of the calendar event."""
        if not reference_object.recurrences:
            return None
        return reference_object.provide_list_in_timezone(reference_object.recurrences.rdates)

    @classmethod
    def value_exrule(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Optional[list[vRecur]]:
        """Return the exrule of the calendar event."""
        if not reference_object.recurrences or not reference_object.recurrences.exrules:
            return None
        return [build_rrule_from_recurrences_rrule(r) for r in reference_object.recurrences.exrules]

    @classmethod
    def value_exdate(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Optional[list[datetime]]:
        """Return the exdate of the calendar event."""
        if not reference_object.recurrences:
            return None
        return reference_object.provide_list_in_timezone(reference_object.recurrences.exdates)

    @classmethod
    def value_unique_id(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> str:
        """Return an unique identifier for an event."""
        if reference_object.amends:
            return cls.value_unique_id(reference_object.amends, request=request)
        return f"{cls.name}-{reference_object.id}"

    @classmethod
    def value_recurrence_id(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> Optional[Union[datetime, date]]:
        """Return the recurrence id of the calendar event."""
        if reference_object.amends:
            return reference_object.amends.value_start_datetime(reference_object, request=request)
        return None

    @classmethod
    def value_color(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ) -> str:
        """Return the color of the calendar."""
        return cls.get_color(request)

    @classmethod
    def value_reference_object(
        cls, reference_object: "CalendarEvent", request: HttpRequest | None = None
    ):
        """Return the reference object itself."""
        return reference_object

    @classmethod
    def get_objects(
        cls, request: HttpRequest | None = None, params: dict[str, any] | None = None
    ) -> Iterable:
        """Return all objects that should be included in the calendar."""
        return cls.objects.instance_of(cls)

    def save(self, *args, **kwargs):
        if (
            self.datetime_start
            and self.datetime_end
            and self.datetime_start.tzinfo != self.datetime_end.tzinfo
        ):
            self.datetime_end = self.datetime_end.astimezone(self.datetime_start.tzinfo)
        if self.datetime_start and self.datetime_end:
            self.timezone = self.datetime_start.tzinfo
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = _("Calendar Event")
        verbose_name_plural = _("Calendar Events")
        constraints = [
            models.CheckConstraint(
                check=~Q(datetime_start__isnull=True, date_start__isnull=True),
                name="datetime_start_or_date_start",
            ),
            models.CheckConstraint(
                check=~Q(datetime_end__isnull=True, date_end__isnull=True),
                name="datetime_end_or_date_end",
            ),
            models.CheckConstraint(
                check=~Q(datetime_start__isnull=False, timezone__isnull=True),
                name="timezone_if_datetime_start",
            ),
            models.CheckConstraint(
                check=~Q(datetime_end__isnull=False, timezone__isnull=True),
                name="timezone_if_datetime_end",
            ),
        ]
        ordering = ["datetime_start", "date_start", "datetime_end", "date_end"]


class FreeBusy(CalendarEvent):
    pass


class BirthdayEvent(CalendarEventMixin):
    """A calendar feed with all birthdays."""

    name = "birthdays"
    verbose_name = _("Birthdays")
    permission_required = "core.view_birthday_calendar"

    @classmethod
    def value_title(cls, reference_object: Person, request: HttpRequest | None = None) -> str:
        return _("{}'s birthday").format(reference_object.addressing_name)

    @classmethod
    def value_description(cls, reference_object: Person, request: HttpRequest | None = None) -> str:
        return ("{name} was born on {birthday}.").format(
            name=reference_object.addressing_name,
            birthday=date_format(reference_object.date_of_birth),
        )

    @classmethod
    def value_start_datetime(
        cls, reference_object: Person, request: HttpRequest | None = None
    ) -> date:
        return reference_object.date_of_birth

    @classmethod
    def value_rrule(cls, reference_object: Person, request: HttpRequest | None = None) -> vRecur:
        return build_rrule_from_text("FREQ=YEARLY")

    @classmethod
    def value_unique_id(cls, reference_object: Person, request: HttpRequest | None = None) -> str:
        return f"birthday-{reference_object.id}"

    @classmethod
    def value_meta(cls, reference_object: Person, request: HttpRequest | None = None) -> dict:
        return {
            "name": reference_object.addressing_name,
            "date_of_birth": reference_object.date_of_birth.isoformat(),
        }

    @classmethod
    def get_color(cls, request: HttpRequest | None = None) -> str:
        return get_site_preferences()["calendar__birthday_color"]

    @classmethod
    def get_objects(
        cls, request: HttpRequest | None = None, params: dict[str, any] | None = None
    ) -> QuerySet:
        qs = Person.objects.filter(date_of_birth__isnull=False)
        if request:
            qs = qs.filter(
                Q(pk=request.user.person.pk)
                | Q(pk__in=get_objects_for_user(request.user, "core.view_personal_details", qs))
            )
        return qs


class Holiday(CalendarEvent):
    """Holiday model for keeping track of school holidays."""

    name = "holidays"
    verbose_name = _("Holidays")
    permission_required = "core.view_holiday_calendar"

    @classmethod
    def value_title(cls, reference_object: "Holiday", request: HttpRequest | None = None) -> str:
        return reference_object.holiday_name

    @classmethod
    def value_description(
        cls, reference_object: "Holiday", request: HttpRequest | None = None
    ) -> str:
        return ""

    @classmethod
    def get_color(cls, request: HttpRequest | None = None) -> str:
        return get_site_preferences()["calendar__holiday_color"]

    objects = PolymorphicBaseManager.from_queryset(HolidayQuerySet)()

    holiday_name = models.CharField(verbose_name=_("Name"), max_length=255)

    def get_days(self) -> Iterator[date]:
        """Get all days included in the holiday."""
        delta = self.date_end - self.date_start
        for i in range(delta.days + 1):
            yield self.date_start + timedelta(days=i)

    @classmethod
    def in_week(cls, week: CalendarWeek) -> dict[int, Optional["Holiday"]]:
        """Get the holidays that are active in a given week."""
        per_weekday = {}
        holidays = Holiday.objects.in_week(week)

        for weekday in range(0, 7):
            holiday_date = week[weekday]
            filtered_holidays = list(
                filter(
                    lambda h: holiday_date >= h.date_start and holiday_date <= h.date_end,
                    holidays,
                )
            )
            if filtered_holidays:
                per_weekday[weekday] = filtered_holidays[0]

        return per_weekday

    @classmethod
    def get_ex_dates(cls, datetime_start, datetime_end, recurrence):
        """Get the dates to exclude for holidays."""
        recurrence.dtstart = recurrence.dtstart.astimezone(timezone.get_current_timezone())
        holiday_dates = [
            h["DTSTART"].dt
            for h in Holiday.get_single_events(start=datetime_start, end=datetime_end)
        ]
        exdates = [
            h.astimezone(timezone.utc)
            for h in recurrence.occurrences()
            if h.date() in holiday_dates
        ]
        return exdates

    def __str__(self) -> str:
        return self.holiday_name

    class Meta:
        verbose_name = _("Holiday")
        verbose_name_plural = _("Holidays")
        permissions = [("view_holiday_calendar", _("Can view holiday calendar"))]


class PersonalEvent(CalendarEvent):
    name = "personal"
    verbose_name = _("Personal events")

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    description = models.TextField(verbose_name=_("Description"), blank=True)
    location = models.CharField(max_length=255, verbose_name=_("Location"), blank=True)

    owner = models.ForeignKey(
        Person,
        related_name="owned_events",
        on_delete=models.CASCADE,
        verbose_name=_("Owner"),
    )
    persons = models.ManyToManyField(Person, related_name="+", blank=True)
    groups = models.ManyToManyField(Group, related_name="+", blank=True)

    @classmethod
    def get_color(cls, request: HttpRequest | None = None) -> str:
        return get_site_preferences()["calendar__personal_event_color"]

    @classmethod
    def value_title(
        cls, reference_object: "PersonalEvent", request: HttpRequest | None = None
    ) -> str:
        """Return the title of the calendar event."""
        return reference_object.title

    @classmethod
    def value_description(
        cls, reference_object: "PersonalEvent", request: HttpRequest | None = None
    ) -> str:
        """Return the description of the calendar event."""
        return reference_object.description

    @classmethod
    def value_location(
        cls, reference_object: "PersonalEvent", request: HttpRequest | None = None
    ) -> str:
        """Return the location of the calendar event."""
        return reference_object.location

    @classmethod
    def value_meta(
        cls, reference_object: "PersonalEvent", request: HttpRequest | None = None
    ) -> dict[str, any]:
        """Get the meta of the event."""

        return {
            "id": reference_object.id,
            "owner": {
                "id": reference_object.owner.pk,
                "full_name": reference_object.owner.full_name,
            },
            "persons": [
                {
                    "id": p.pk,
                    "full_name": p.full_name,
                }
                for p in reference_object.persons.all()
            ],
            "groups": [
                {"id": g.pk, "name": g.name, "short_name": g.short_name}
                for g in reference_object.groups.all()
            ],
            "description": reference_object.description,
            "recurrences": str(reference_object.recurrences),
            "can_edit": request.user.has_perm("core.edit_personal_event_rule", reference_object),
            "can_delete": request.user.has_perm(
                "core.delete_personal_event_rule", reference_object
            ),
        }

    @classmethod
    def get_objects(
        cls, request: HttpRequest | None = None, params: dict[str, any] | None = None
    ) -> QuerySet:
        qs = super().get_objects(request=request, params=params)
        if request:
            qs = qs.filter(
                Q(owner=request.user.person)
                | Q(persons=request.user.person)
                | Q(groups__members=request.user.person)
            ).distinct()
        return qs


class Organisation(ExtensibleModel):
    name = models.CharField(verbose_name=_("Name"), max_length=255)
    email = models.CharField(verbose_name=_("Email"), max_length=255)

    related_group = models.ForeignKey(
        Group, on_delete=models.CASCADE, verbose_name=_("Related group")
    )

    def __str__(self) -> str:
        return f"{self.name}"

    class Meta:
        verbose_name = _("Organisation")
        verbose_name_plural = _("Organisations")



# Generated by Django 4.1.9 on 2023-07-06 21:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0049_oauthapplication_post_logout_redirect_uris"),
    ]

    operations = [
        migrations.AddField(
            model_name="activity",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="additionalfield",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="announcement",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="announcementrecipient",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="custommenu",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="custommenuitem",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="dashboardwidgetorder",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="datacheckresult",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="group",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="grouptype",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="notification",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="pdffile",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="person",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="persongroupthrough",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="room",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="schoolterm",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
        migrations.AddField(
            model_name="taskuserassignment",
            name="managed_by_app_label",
            field=models.CharField(
                blank=True,
                editable=False,
                max_length=255,
                verbose_name="App label of app responsible for managing this instance",
            ),
        ),
    ]



from django.utils.safestring import mark_safe

from django_tables2 import CheckBoxColumn
from django_tables2.utils import A, AttributeDict, computed_values


class MaterializeCheckboxColumn(CheckBoxColumn):
    """Checkbox column with Materialize support."""

    empty_values = ()

    @property
    def header(self):
        """Render the header cell."""
        default = {"type": "checkbox"}
        general = self.attrs.get("input")
        specific = self.attrs.get("th__input")
        attrs = AttributeDict(default, **(specific or general or {}))
        return mark_safe("<label><input %s/><span></span></label>" % attrs.as_html())  # noqa

    def render(self, value, bound_column, record):
        """Render a data cell."""
        default = {"type": "checkbox", "name": bound_column.name, "value": value}
        if self.is_checked(value, record):
            default.update({"checked": "checked"})

        general = self.attrs.get("input")
        specific = self.attrs.get("td__input")

        attrs = dict(default, **(specific or general or {}))
        attrs = computed_values(attrs, kwargs={"record": record, "value": value})
        return mark_safe(  # noqa
            "<label><input %s/><span></span</label>" % AttributeDict(attrs).as_html()
        )


class SelectColumn(MaterializeCheckboxColumn):
    """Column with a check box prepared for `ActionForm` forms."""

    def __init__(self, *args, **kwargs):
        kwargs["attrs"] = {
            "td__input": {"name": "selected_objects"},
            "th__input": {"class": "select--header-box"},
        }
        kwargs.setdefault("accessor", A("pk"))
        super().__init__(*args, **kwargs)



# Generated by Django 3.2.13 on 2022-06-05 10:55

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0040_oauth_allowed_scopes_max_length_255'),
    ]

    operations = [
        migrations.AlterField(
            model_name='person',
            name='sex',
            field=models.CharField(blank=True, choices=[('f', 'female'), ('m', 'male'), ('x', 'other')], max_length=1, verbose_name='Sex'),
        ),
    ]



from typing import Iterable

from django.utils import timezone

import graphene
from graphene_django import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
)

from ..models import PersonalEvent
from .base import (
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
)


class PersonalEventType(DjangoObjectType):
    class Meta:
        model = PersonalEvent
        fields = (
            "id",
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "owner",
            "persons",
            "groups",
        )

    recurrences = graphene.String()


class PersonalEventBatchCreateMutation(PermissionBatchPatchMixin, DjangoBatchCreateMutation):
    class Meta:
        model = PersonalEvent
        permissions = ("core.create_personal_event_with_invitations_rule",)
        only_fields = (
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "recurrences",
            "persons",
            "groups",
        )
        field_types = {"recurrences": graphene.String(), "location": graphene.String()}

    @classmethod
    def get_permissions(cls, root, info, input) -> Iterable[str]:  # noqa
        if [len(event.persons) == 0 and len(event.groups) == 0 for event in input].all():
            return ("core.create_personal_event_rule",)
        return cls._meta.permissions

    @classmethod
    def before_mutate(cls, root, info, input):  # noqa
        for event in input:
            event["owner"] = info.context.user.person.id
        return input

    @classmethod
    def handle_datetime_start(cls, value, name, info):
        value = value.replace(tzinfo=timezone.get_default_timezone())
        return value

    @classmethod
    def handle_datetime_end(cls, value, name, info):
        value = value.replace(tzinfo=timezone.get_default_timezone())
        return value


class PersonalEventBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = PersonalEvent
        permissions = ("core.delete_personal_event_rule",)


class PersonalEventBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = PersonalEvent
        permissions = ("core.change_personalevent",)
        only_fields = (
            "id",
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "recurrences",
            "persons",
            "groups",
        )
        field_types = {"recurrences": graphene.String(), "location": graphene.String()}

    @classmethod
    def get_permissions(cls, root, info, input, id, obj) -> Iterable[str]:  # noqa
        if info.context.user.has_perm("core.edit_personal_event_rule", obj):
            return []
        return cls._meta.permissions

    @classmethod
    def handle_datetime_start(cls, value, name, info):
        value = value.replace(tzinfo=timezone.get_default_timezone())
        return value

    @classmethod
    def handle_datetime_end(cls, value, name, info):
        value = value.replace(tzinfo=timezone.get_default_timezone())
        return value



import pytest

from aleksis.core.mixins import RegistryObject


def test_registry_object_name():
    class ExampleRegistry(RegistryObject):
        pass

    class ExampleWithManualName(ExampleRegistry):
        name = "example_a"

    class ExampleWithAutomaticName(ExampleRegistry):
        pass

    class ExampleWithOverridenAutomaticName(ExampleWithManualName):
        name = "example_b"

    class ExampleWithOverridenManualName(ExampleWithAutomaticName):
        name = "example_bb"

    assert ExampleRegistry.name == ""
    assert ExampleWithManualName.name == "example_a"
    assert ExampleWithAutomaticName.name == "ExampleWithAutomaticName"
    assert ExampleWithOverridenAutomaticName.name == "example_b"
    assert ExampleWithOverridenManualName.name == "example_bb"


def test_registry_object_registry():
    class ExampleRegistry(RegistryObject):
        pass

    class ExampleA(ExampleRegistry):
        pass

    class ExampleB(ExampleRegistry):
        pass

    class ExampleAA(ExampleA):
        name = "example_aa"

    assert ExampleRegistry.registered_objects_dict == {
        "ExampleA": ExampleA,
        "ExampleB": ExampleB,
        "example_aa": ExampleAA,
    }
    assert ExampleRegistry.registered_objects_dict == ExampleRegistry._registry

    assert ExampleRegistry.registered_objects_list == [
        ExampleA,
        ExampleB,
        ExampleAA,
    ]

    assert ExampleRegistry.get_object_by_name("ExampleA") == ExampleA
    assert ExampleRegistry.get_object_by_name("ExampleB") == ExampleB
    assert ExampleRegistry.get_object_by_name("example_aa") == ExampleAA



"""Helpers/overrides for django-allauth."""

from typing import Any, Optional

from django.conf import settings
from django.contrib.auth.validators import ASCIIUsernameValidator
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _

from allauth.account.adapter import DefaultAccountAdapter
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from oauth2_provider.models import AbstractApplication
from oauth2_provider.oauth2_validators import OAuth2Validator
from oauth2_provider.scopes import BaseScopes
from oauth2_provider.views.mixins import (
    ClientProtectedResourceMixin as _ClientProtectedResourceMixin,
)
from oauthlib.common import Request as OauthlibRequest

from .apps import AppConfig
from .core_helpers import get_site_preferences


class OurSocialAccountAdapter(DefaultSocialAccountAdapter):
    """Customised adapter that recognises other authentication mechanisms."""

    def validate_disconnect(self, account, accounts):
        """Validate whether or not the socialaccount account can be safely disconnected.

        Honours other authentication backends, i.e. ignores unusable passwords if LDAP is used.
        """
        if "django_auth_ldap.backend.LDAPBackend" in settings.AUTHENTICATION_BACKENDS:
            # Ignore upstream validation error as we do not need a usable password
            return None

        # Let upstream decide whether we can disconnect or not
        return super().validate_disconnect(account, accounts)


class OurAccountAdapter(DefaultAccountAdapter):
    """Customised adapter to allow to disable signup."""

    def is_open_for_signup(self, request):
        return get_site_preferences()["auth__signup_enabled"]


class CustomOAuth2Validator(OAuth2Validator):
    def get_additional_claims(self, request: OauthlibRequest) -> dict[str, Any]:
        # Pull together scopes from request and from access token
        scopes = request.scopes.copy()
        if request.access_token:
            scopes += request.access_token.scope.split(" ")

        claims = {}
        # Pull together claim data from all apps
        for app in AppConfig.__subclasses__():
            claims.update(app.get_additional_claims(scopes, request))

        return claims


class AppScopes(BaseScopes):
    """Scopes backend for django-oauth-toolkit gathering scopes from apps.

    Will call the respective method on all known AlekSIS app configs and
    join the results.
    """

    def get_all_scopes(self) -> dict[str, str]:
        scopes = {}
        for app in AppConfig.__subclasses__():
            scopes |= app.get_all_scopes()
        return scopes

    def get_available_scopes(
        self,
        application: Optional[AbstractApplication] = None,
        request: Optional[HttpRequest] = None,
        *args,
        **kwargs,
    ) -> list[str]:
        scopes = []
        for app in AppConfig.__subclasses__():
            scopes += app.get_available_scopes()
        # Filter by allowed scopes of requesting application
        if application and application.allowed_scopes:
            scopes = list(filter(lambda scope: scope in application.allowed_scopes, scopes))
        return scopes

    def get_default_scopes(
        self,
        application: Optional[AbstractApplication] = None,
        request: Optional[HttpRequest] = None,
        *args,
        **kwargs,
    ) -> list[str]:
        scopes = []
        for app in AppConfig.__subclasses__():
            scopes += app.get_default_scopes()
        # Filter by allowed scopes of requesting application
        if application and application.allowed_scopes:
            scopes = list(filter(lambda scope: scope in application.allowed_scopes, scopes))
        return scopes


class ClientProtectedResourceMixin(_ClientProtectedResourceMixin):
    """Mixin for protecting resources with client authentication as mentioned in rfc:`3.2.1`.

    This involves authenticating with any of: HTTP Basic Auth, Client Credentials and
    Access token in that order. Breaks off after first validation.

    This sub-class extends the functionality of Django OAuth Toolkit's mixin with support
    for AlekSIS's `allowed_scopes` feature. For applications that have configured allowed
    scopes, the required scopes for the view are checked to be a subset of the application's
    allowed scopes (best to be combined with ScopedResourceMixin).
    """

    def authenticate_client(self, request: HttpRequest) -> bool:
        """Return a boolean representing if client is authenticated with client credentials.

        If the view has configured required scopes, they are verified against the application's
        allowed scopes.
        """
        # Build an OAuth request so we can handle client information
        core = self.get_oauthlib_core()
        uri, http_method, body, headers = core._extract_params(request)
        oauth_request = OauthlibRequest(uri, http_method, body, headers)

        # Verify general authentication of the client
        if not core.server.request_validator.authenticate_client(oauth_request):
            # Client credentials were invalid
            return False

        # Verify scopes of configured application
        # The OAuth request was enriched with a reference to the Application when using the
        #  validator above.
        if not oauth_request.client.allowed_scopes:
            # If there are no allowed scopes, the client is not allowed to access this resource
            return False

        required_scopes = set(self.get_scopes() or [])
        allowed_scopes = set(AppScopes().get_available_scopes(oauth_request.client) or [])
        return required_scopes.issubset(allowed_scopes)


def validate_username_preference_regex(value: str):
    regex = get_site_preferences()["auth__allowed_username_regex"]
    return RegexValidator(regex)(value)


def validate_username_preference_disallowed_uid(value: str):
    if value in get_site_preferences()["auth__disallowed_uids"].split(","):
        raise ValidationError(_("This username is not allowed."))


custom_username_validators = [
    validate_username_preference_regex,
    ASCIIUsernameValidator(),
    validate_username_preference_disallowed_uid,
]



from django.core.exceptions import PermissionDenied, ValidationError
from django.utils.translation import gettext as _

from graphene_django import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
)

from ..models import SchoolTerm
from .base import (
    DjangoFilterMixin,
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
    PermissionsTypeMixin,
)


class SchoolTermType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = SchoolTerm
        filter_fields = {
            "name": ["icontains", "exact"],
            "date_start": ["exact", "lt", "lte", "gt", "gte"],
            "date_end": ["exact", "lt", "lte", "gt", "gte"],
        }
        fields = ("id", "name", "date_start", "date_end")

    @classmethod
    def get_queryset(cls, queryset, info, **kwargs):
        if not info.context.user.has_perm("core.view_schoolterm_rule"):
            raise PermissionDenied

        return queryset  # FIXME filter this queryset based on permissions


class SchoolTermBatchCreateMutation(PermissionBatchPatchMixin, DjangoBatchCreateMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.create_school_term",)  # FIXME
        only_fields = ("id", "name", "date_start", "date_end")

    @classmethod
    def validate(cls, root, info, input, inputs):  # noqa
        for input in inputs:  # noqa
            date_start = input.get("date_start")
            date_end = input.get("date_end")
            if date_end < date_start:
                raise ValidationError(_("The start date must be earlier than the end date."))

            qs = SchoolTerm.objects.within_dates(date_start, date_end)
            if qs.exists():
                raise ValidationError(
                    _("There is already a school term for this time or a part of this time.")
                )


class SchoolTermBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.delete_school_term",)


class SchoolTermBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.change_school_term",)  # FIXME
        only_fields = ("id", "name", "date_start", "date_end")



# Generated by Django 4.2.7 on 2023-12-09 11:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0059_drop_site'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='person',
            name='unique_short_name',
        ),
        migrations.AlterField(
            model_name='person',
            name='short_name',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Short name'),
            preserve_default=False,
        ),
        migrations.AddConstraint(
            model_name='person',
            constraint=models.UniqueConstraint(condition=models.Q(('short_name', ''), _negated=True), fields=('short_name',), name='unique_short_name'),
        ),
        migrations.AddConstraint(
            model_name='person',
            constraint=models.UniqueConstraint(condition=models.Q(('email', ''), _negated=True), fields=('email',), name='unique_email'),
        ),
    ]



# Generated by Django 4.1.9 on 2023-06-17 10:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0048_delete_personalicalurl"),
    ]

    operations = [
        migrations.AddField(
            model_name="oauthapplication",
            name="post_logout_redirect_uris",
            field=models.TextField(
                blank=True, help_text="Allowed Post Logout URIs list, space separated"
            ),
        ),
    ]



# Generated by Django 3.2.8 on 2021-11-04 19:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0023_oauth_application_model'),
        ('oauth2_provider', '0004_auto_20200902_2022'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oauthapplication',
            name='authorization_grant_type',
            field=models.CharField(blank=True, choices=[('authorization-code', 'Authorization code'), ('implicit', 'Implicit'), ('password', 'Resource owner password-based'), ('client-credentials', 'Client credentials'), ('openid-hybrid', 'OpenID connect hybrid')], max_length=32, null=True),
        ),
    ]



from django.templatetags.static import static

import graphene


class ThemeLogoType(graphene.ObjectType):
    url = graphene.String(required=False)


class SitePreferencesType(graphene.ObjectType):
    general_title = graphene.String()

    theme_logo = graphene.Field(ThemeLogoType)
    theme_primary = graphene.String()
    theme_secondary = graphene.String()

    footer_imprint_url = graphene.String()
    footer_privacy_url = graphene.String()

    account_person_prefer_photo = graphene.Boolean()

    def resolve_general_title(parent, info, **kwargs):
        return parent["general__title"]

    def resolve_theme_logo(parent, info, **kwargs):
        return (
            parent["theme__logo"]
            if parent["theme__logo"]
            else {"url": static("/img/aleksis-banner.svg")}
        )

    def resolve_theme_primary(parent, info, **kwargs):
        return parent["theme__primary"]

    def resolve_theme_secondary(parent, info, **kwargs):
        return parent["theme__secondary"]

    def resolve_footer_imprint_url(parent, info, **kwargs):
        return parent["footer__imprint_url"]

    def resolve_footer_privacy_url(parent, info, **kwargs):
        return parent["footer__privacy_url"]

    def resolve_account_person_prefer_photo(parent, info, **kwargs):
        return parent["account__person_prefer_photo"]



# Generated by Django 3.2.16 on 2022-12-16 11:50

from django.db import migrations, models
import django.utils.timezone
import oauth2_provider.generators
import oauth2_provider.models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0044_task_assignment_result_fetched'),
    ]

    operations = [
        migrations.RenameField(
            model_name='datacheckresult',
            old_name='check',
            new_name='data_check',
        ),
    ]



# noqa

from django.contrib import admin

from guardian.admin import GuardedModelAdminMixin
from reversion.admin import VersionAdmin

from .mixins import BaseModelAdmin
from .models import (
    Activity,
    Announcement,
    AnnouncementRecipient,
    CustomMenuItem,
    DataCheckResult,
    Group,
    Notification,
    Organisation,
    Person,
)

admin.site.register(Activity, VersionAdmin)
admin.site.register(Notification, VersionAdmin)
admin.site.register(CustomMenuItem, VersionAdmin)


class AnnouncementRecipientInline(admin.StackedInline):
    model = AnnouncementRecipient


class AnnouncementAdmin(BaseModelAdmin, VersionAdmin):
    inlines = [
        AnnouncementRecipientInline,
    ]


class GuardedVersionAdmin(GuardedModelAdminMixin, VersionAdmin):
    pass


admin.site.register(Announcement, AnnouncementAdmin)
admin.site.register(DataCheckResult)
admin.site.register(Person, GuardedVersionAdmin)
admin.site.register(Group, GuardedVersionAdmin)
admin.site.register(Organisation)
